# hg.py - repository classes for mercurial
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
# Copyright 2006 Vadim Gelfer <vadim.gelfer@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import os
import stat

from .i18n import _
from .node import (
    short,
)

from . import (
    error,
    localrepo,
    lock,
    merge as mergemod,
    scmutil,
    util,
    verify as verifymod,
)
from .cmd_impls import (
    clone as clone_impl,
    incoming as inc_impl,
    outgoing as out_impl,
    update as up_impl,
)
from .repo import factory as repo_factory
from .utils import (
    urlutil,
)


release = lock.release


def addbranchrevs(lrepo, *args, **kwargs):
    msg = b'``hg.addbranchrevs(...)` moved to `urlutil.add_branch_revs(...)`'
    lrepo.ui.deprecwarn(msg, b'7.3')
    return urlutil.add_branch_revs(lrepo, *args, **kwargs)


# a list of (ui, repo) functions called for wire peer initialization
wirepeersetupfuncs = repo_factory.wirepeersetupfuncs
repo_schemes = repo_factory.repo_schemes
peer_schemes = repo_factory.peer_schemes


def islocal(repo: bytes) -> bool:
    util.nouideprecwarn(
        b'hg.islocal(path), moved to repo.factor.is_local(path)', b'7.3'
    )
    return repo_factory.is_local(repo)


def openpath(*args, **kwargs):
    msg = b'`hg.openpath(...)` moved to scmutil.open_path(...)'
    util.nouideprecwarn(msg, b'7.3')
    return scmutil.open_path(*args, **kwargs)


def peer(uiorrepo, *args, **kwargs):
    ui = getattr(uiorrepo, 'ui', uiorrepo)
    msg = b'``hg.peer(...)` moved to `repo.factory.peer(...)`'
    ui.deprecwarn(msg, b'7.3')
    return repo_factory.peer(uiorrepo, *args, **kwargs)


def repository(uiorrepo, *args, **kwargs):
    ui = getattr(uiorrepo, 'ui', uiorrepo)
    msg = b'``hg.repository(...)` moved to `repo.factory.repository(...)`'
    ui.deprecwarn(msg, b'7.3')
    return repo_factory.repository(uiorrepo, *args, **kwargs)


def defaultdest(source):
    msg = b'hg.defaultdest(src), moved to cmd_impls.clone.default_dest(src)'
    util.nouideprecwarn(msg, b'7.3')
    return clone_impl.default_dest(source)


def share(ui, *args, **kwargs):
    msg = b'``hg.share(...)` moved to `cmd_impls.clone.share(...)`'
    ui.deprecwarn(msg, b'7.3')
    return clone_impl.share(ui, *args, **kwargs)


def unshare(ui, *args, **kwargs):
    msg = b'``hg.unshare(...)` moved to `cmd_impls.clone.unshare(...)`'
    ui.deprecwarn(msg, b'7.3')
    return clone_impl.unshare(ui, *args, **kwargs)


def clone(ui, *args, **kwargs):
    msg = b'``hg.clone(...)` moved to `cmd_impls.clone.clone(...)`'
    ui.deprecwarn(msg, b'7.3')
    return clone_impl.clone(ui, *args, **kwargs)


def sharedreposource(repo):
    """Returns repository object for source repository of a shared repo.

    If repo is not a shared repository, returns None.
    """
    if repo.sharedpath == repo.path:
        return None

    if hasattr(repo, 'srcrepo') and repo.srcrepo:
        return repo.srcrepo

    # the sharedpath always ends in the .hg; we want the path to the repo
    source = repo.vfs.split(repo.sharedpath)[0]
    srcurl, branches = urlutil.parseurl(source)
    srcrepo = repo_factory.repository(repo.ui, srcurl)
    repo.srcrepo = srcrepo
    return srcrepo


# Recomputing caches is often slow on big repos, so copy them.
def _copycache(srcrepo, dstcachedir, fname):
    """copy a cache from srcrepo to destcachedir (if it exists)"""
    srcfname = srcrepo.cachevfs.join(fname)
    dstfname = os.path.join(dstcachedir, fname)
    if os.path.exists(srcfname):
        if not os.path.exists(dstcachedir):
            os.mkdir(dstcachedir)
        util.copyfile(srcfname, dstfname)


def updaterepo(repo, node, overwrite, updatecheck=None):
    """Update the working directory to node.

    When overwrite is set, changes are clobbered, merged else

    returns stats (see pydoc mercurial.merge_utils.update.apply_updates)"""
    repo.ui.deprecwarn(
        b'prefer merge.update() or merge.clean_update() over hg.updaterepo()',
        b'5.7',
    )
    return mergemod._update(
        repo,
        node,
        branchmerge=False,
        force=overwrite,
        labels=[b'working copy', b'destination'],
        updatecheck=updatecheck,
    )


def update(repo, *args, **kwargs):
    msg = b'``hg.update(...)` moved to `cmd_impls.update.update(...)`'
    repo.ui.deprecwarn(msg, b'7.3')
    return up_impl.update(repo, *args, **kwargs)


def updatetotally(ui, *args, **kwargs):
    msg = b'``hg.updatetotally` moved to `cmd_impls.update.update_totally`'
    ui.deprecwarn(msg, b'7.3')
    return up_impl.update_totally(ui, *args, **kwargs)


def clean(repo, *args, **kwargs):
    msg = b'``hg.clean(...)` moved to `cmd_impls.update.clean(...)`'
    repo.ui.deprecwarn(msg, b'7.3')
    return up_impl.clean(repo, *args, **kwargs)


def merge(ctx, *args, **kwargs):
    msg = b'``hg.merge(...)` moved to `cmd_impls.update.merge(...)`'
    ctx.repo().ui.deprecwarn(msg, b'7.3')
    return up_impl.merge(ctx, *args, **kwargs)


def abortmerge(ui, *args, **kwargs):
    msg = b'``hg.abortmerge(...)` moved to `cmd_impls.update.abortmerge(...)`'
    ui.deprecwarn(msg, b'7.3')
    return up_impl.abort_merge(ui, *args, **kwargs)


def outgoing(ui, *args, **kwargs):
    msg = b'``hg.outgoing(...)` moved to `cmd_impls.outgoing.outgoing(...)`'
    ui.deprecwarn(msg, b'7.3')
    return out_impl.outgoing(ui, *args, **kwargs)


def incoming(ui, *args, **kwargs):
    msg = b'``hg.incoming(...)` moved to `cmd_impls.outgoing.incoming(...)`'
    ui.deprecwarn(msg, b'7.3')
    return inc_impl.incoming(ui, *args, **kwargs)


def verify(repo, level=None):
    """verify the consistency of a repository"""
    ret = verifymod.verify(repo, level=level)

    # Broken subrepo references in hidden csets don't seem worth worrying about,
    # since they can't be pushed/pulled, and --hidden can be used if they are a
    # concern.

    # pathto() is needed for -R case
    revs = repo.revs(
        b"filelog(%s)", util.pathto(repo.root, repo.getcwd(), b'.hgsubstate')
    )

    if revs:
        repo.ui.status(_(b'checking subrepo links\n'))
        for rev in revs:
            ctx = repo[rev]
            try:
                for subpath in ctx.substate:
                    try:
                        ret = (
                            ctx.sub(subpath, allowcreate=False).verify() or ret
                        )
                    except error.RepoError as e:
                        repo.ui.warn(b'%d: %s\n' % (rev, e))
            except Exception:
                repo.ui.warn(
                    _(b'.hgsubstate is corrupt in revision %s\n')
                    % short(ctx.node())
                )

    return ret


# Files of interest
# Used to check if the repository has changed looking at mtime and size of
# these files.
foi: list[tuple[str, bytes]] = [
    ('spath', b'00changelog.i'),
    ('spath', b'phaseroots'),  # ! phase can change content at the same size
    ('spath', b'obsstore'),
    ('path', b'bookmarks'),  # ! bookmark can change content at the same size
]


class cachedlocalrepo:
    """Holds a localrepository that can be cached and reused."""

    def __init__(self, repo):
        """Create a new cached repo from an existing repo.

        We assume the passed in repo was recently created. If the
        repo has changed between when it was created and when it was
        turned into a cache, it may not refresh properly.
        """
        assert isinstance(repo, localrepo.localrepository)
        self._repo = repo
        self._state, self.mtime = self._repostate()
        self._filtername = repo.filtername

    def fetch(self):
        """Refresh (if necessary) and return a repository.

        If the cached instance is out of date, it will be recreated
        automatically and returned.

        Returns a tuple of the repo and a boolean indicating whether a new
        repo instance was created.
        """
        # We compare the mtimes and sizes of some well-known files to
        # determine if the repo changed. This is not precise, as mtimes
        # are susceptible to clock skew and imprecise filesystems and
        # file content can change while maintaining the same size.

        state, mtime = self._repostate()
        if state == self._state:
            return self._repo, False

        repo = repo_factory.repository(self._repo.baseui, self._repo.url())
        if self._filtername:
            self._repo = repo.filtered(self._filtername)
        else:
            self._repo = repo.unfiltered()
        self._state = state
        self.mtime = mtime

        return self._repo, True

    def _repostate(self):
        state = []
        maxmtime = -1
        for attr, fname in foi:
            prefix = getattr(self._repo, attr)
            p = os.path.join(prefix, fname)
            try:
                st = os.stat(p)
            except OSError:
                st = os.stat(prefix)
            state.append((st[stat.ST_MTIME], st.st_size))
            maxmtime = max(maxmtime, st[stat.ST_MTIME])

        return tuple(state), maxmtime

    def copy(self):
        """Obtain a copy of this class instance.

        A new localrepository instance is obtained. The new instance should be
        completely independent of the original.
        """
        repo = repo_factory.repository(self._repo.baseui, self._repo.origroot)
        if self._filtername:
            repo = repo.filtered(self._filtername)
        else:
            repo = repo.unfiltered()
        c = cachedlocalrepo(repo)
        c._state = self._state
        c.mtime = self.mtime
        return c



# changegroup.py - Mercurial changegroup manipulation functions
#
#  Copyright 2006 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import abc
import enum
import io
import os
import struct
import weakref

from typing import (
    Callable,
    Collection,
    Iterator,
    Optional,
    TYPE_CHECKING,
    cast,
)

from .i18n import _
from .node import (
    hex,
    nullrev,
    short,
)
from .interfaces.types import (
    MatcherT,
    NeedsTypeHint,
    NodeIdT,
    OutboundRevisionT,
    RepoT,
    RevnumT,
)

from .thirdparty import attr
from . import (
    error,
    match as matchmod,
    mdiff,
    phases,
    pycompat,
    requirements,
    revlogutils,
    scmutil,
    util,
)

from .interfaces import (
    changegroup as i_cg,
    compression as i_comp,
    repository as i_repo,
)
from .revlogutils import sidedata as sidedatamod
from .revlogutils import constants as revlog_constants
from .utils import (
    storageutil,
    urlutil,
)

# Force pytype to use the non-vendored package
if TYPE_CHECKING:
    # noinspection PyPackageRequirements
    import attr

# small type alias to help type some function
LookupFnT = Callable[[NodeIdT], NodeIdT]

_CHANGEGROUPV1_DELTA_HEADER = struct.Struct(b"20s20s20s20s")
_CHANGEGROUPV2_DELTA_HEADER = struct.Struct(b"20s20s20s20s20s")
# node p1node p2node basenode linknode flags
_CHANGEGROUPV3_DELTA_HEADER = struct.Struct(b">20s20s20s20s20sH")
# node
# p1node
# p2node
# basenode
# linknode
# flags
# snapshot_level
# raw_size
# compression
# delta_flag
# storage_base_node
# storage_snapshot_level
_CHANGEGROUPV4_DELTA_HEADER = struct.Struct(b">20s20s20s20s20sHbIBB20sb")
_CHANGEGROUPV5_DELTA_HEADER = struct.Struct(b">B20s20s20s20s20sH")

LFS_REQUIREMENT = b'lfs'

readexactly = util.readexactly


def getchunk(stream):
    """return the next chunk from stream as a string"""
    d = readexactly(stream, 4)
    l = struct.unpack(b">l", d)[0]
    if l <= 4:
        if l:
            raise error.Abort(_(b"invalid chunk length %d") % l)
        return b""
    return readexactly(stream, l - 4)


def chunkheader(length):
    """return a changegroup chunk header (string)"""
    return struct.pack(b">l", length + 4)


def closechunk():
    """return a changegroup chunk header (string) for a zero-length chunk"""
    return struct.pack(b">l", 0)


def _fileheader(path):
    """Obtain a changegroup chunk header for a named path."""
    return chunkheader(len(path)) + path


def writechunks(ui, chunks, filename, vfs=None):
    """Write chunks to a file and return its filename.

    The stream is assumed to be a bundle file.
    Existing files will not be overwritten.
    If no filename is specified, a temporary file is created.
    """
    fh = None
    cleanup = None
    try:
        if filename:
            if vfs:
                fh = vfs.open(filename, b"wb")
            else:
                # Increase default buffer size because default is usually
                # small (4k is common on Linux).
                fh = open(filename, "wb", 131072)
        else:
            fd, filename = pycompat.mkstemp(prefix=b"hg-bundle-", suffix=b".hg")
            fh = os.fdopen(fd, "wb")
        cleanup = filename
        for c in chunks:
            fh.write(c)
        cleanup = None
        return filename
    finally:
        if fh is not None:
            fh.close()
        if cleanup is not None:
            if filename and vfs:
                vfs.unlink(cleanup)
            else:
                os.unlink(cleanup)


def _dbg_ubdl_line(
    ui,
    indent,
    key,
    base_value=None,
    percentage_base=None,
    percentage_key=None,
):
    """Print one line of debug_unbundle_debug_info"""
    line = b"DEBUG-UNBUNDLING: "
    line += b' ' * (2 * indent)
    key += b":"
    padding = b''
    if base_value is not None:
        assert len(key) + 1 + (2 * indent) <= _KEY_PART_WIDTH
        line += key.ljust(_KEY_PART_WIDTH - (2 * indent))
        if isinstance(base_value, float):
            line += b"%14.3f seconds" % base_value
        else:
            line += b"%10d" % base_value
            padding = b'            '
    else:
        line += key

    if percentage_base is not None:
        line += padding
        padding = b''
        assert base_value is not None
        percentage = base_value * 100 // percentage_base
        if percentage_key is not None:
            line += b" (%3d%% of %s)" % (
                percentage,
                percentage_key,
            )
        else:
            line += b" (%3d%%)" % percentage

    line += b'\n'
    ui.write_err(line)


def _sumf(items):
    # python < 3.8 does not support a `start=0.0` argument to sum
    # So we have to cheat a bit until we drop support for those version
    if not items:
        return 0.0
    return sum(items)


def display_unbundle_debug_info(ui, debug_info):
    """display an unbundling report from debug information"""
    cl_info = []
    mn_info = []
    fl_info = []
    _dispatch = [
        (b'CHANGELOG:', cl_info),
        (b'MANIFESTLOG:', mn_info),
        (b'FILELOG:', fl_info),
    ]
    for e in debug_info:
        for prefix, info in _dispatch:
            if e["target-revlog"].startswith(prefix):
                info.append(e)
                break
        else:
            assert False, 'unreachable'
    each_info = [
        (b'changelog', cl_info),
        (b'manifests', mn_info),
        (b'files', fl_info),
    ]

    # General Revision Countss
    _dbg_ubdl_line(ui, 0, b'revisions', len(debug_info))
    for key, info in each_info:
        if not info:
            continue
        _dbg_ubdl_line(ui, 1, key, len(info), len(debug_info))

    # General Time spent
    all_durations = [e['duration'] for e in debug_info]
    all_durations.sort()
    total_duration = _sumf(all_durations)
    _dbg_ubdl_line(ui, 0, b'total-time', total_duration)

    for key, info in each_info:
        if not info:
            continue
        durations = [e['duration'] for e in info]
        durations.sort()
        _dbg_ubdl_line(ui, 1, key, _sumf(durations), total_duration)

    # Count and cache reuse per delta types
    each_types = {}
    for key, info in each_info:
        each_types[key] = types = {
            b'full': 0,
            b'full-cached': 0,
            b'snapshot': 0,
            b'snapshot-cached': 0,
            b'delta': 0,
            b'delta-cached': 0,
            b'unknown': 0,
            b'unknown-cached': 0,
        }
        for e in info:
            types[e['type']] += 1
            if e['using-cached-base']:
                types[e['type'] + b'-cached'] += 1

    EXPECTED_TYPES = (b'full', b'snapshot', b'delta', b'unknown')
    if debug_info:
        _dbg_ubdl_line(ui, 0, b'type-count')
    for key, info in each_info:
        if info:
            _dbg_ubdl_line(ui, 1, key)
        t = each_types[key]
        for tn in EXPECTED_TYPES:
            if t[tn]:
                tc = tn + b'-cached'
                _dbg_ubdl_line(ui, 2, tn, t[tn])
                _dbg_ubdl_line(ui, 3, b'cached', t[tc], t[tn])

    # time perf delta types and reuse
    each_type_time = {}
    for key, info in each_info:
        each_type_time[key] = t = {
            b'full': [],
            b'full-cached': [],
            b'snapshot': [],
            b'snapshot-cached': [],
            b'delta': [],
            b'delta-cached': [],
            b'unknown': [],
            b'unknown-cached': [],
        }
        for e in info:
            t[e['type']].append(e['duration'])
            if e['using-cached-base']:
                t[e['type'] + b'-cached'].append(e['duration'])
        for t_key, value in list(t.items()):
            value.sort()
            t[t_key] = _sumf(value)

    if debug_info:
        _dbg_ubdl_line(ui, 0, b'type-time')
    for key, info in each_info:
        if info:
            _dbg_ubdl_line(ui, 1, key)
        t = each_type_time[key]
        td = total_duration  # to same space on next lines
        for tn in EXPECTED_TYPES:
            if t[tn]:
                tc = tn + b'-cached'
                _dbg_ubdl_line(ui, 2, tn, t[tn], td, b"total")
                _dbg_ubdl_line(ui, 3, b'cached', t[tc], td, b"total")


_NAME_TO_REVLOG_COMP = {
    b'none': i_comp.REVLOG_COMP_NONE,
    b'zlib': i_comp.REVLOG_COMP_ZLIB,
    b'zstd': i_comp.REVLOG_COMP_ZSTD,
}


class WireDeltaCompression(enum.IntEnum):
    """encode the compression of a delta over the wire

    Note that NO_COMPRESSION ≠ UNCOMPRESSED.

    NO_COMPRESSION means the incoming delta is uncompressed and we don't have
    any information about the source compression.

    UNCOMPRESSED means the incoming delta is "compressed" with the "none"
    compression, with the associated impact on the delta content. It usually
    imply the delta wasn't compressed in the source storage.
    """

    NO_COMPRESSION = 0
    UNCOMPRESSED = 1
    ZLIB_COMPRESSED = 2
    ZSTD_COMPRESSED = 3

    @staticmethod
    def accepted(
        caps: Collection[bytes] | None,
    ) -> frozenset[i_comp.RevlogCompHeader]:
        """return accepted compression from a bundle capability set"""
        if caps is None:
            return frozenset()
        b2caps = urlutil.b2_caps_from_bundle_caps(caps)
        cap_set = b2caps.get(b'delta-compression', ())
        accepted = set()
        for cap in cap_set:
            if (rl_cmp := _NAME_TO_REVLOG_COMP.get(cap)) is not None:
                accepted.add(rl_cmp)
        return frozenset(accepted)

    @staticmethod
    def filter_known(
        caps: Collection[bytes] | None,
    ) -> set[bytes]:
        if caps is None:
            return set()
        b2caps = urlutil.b2_caps_from_bundle_caps(caps)
        cap_set = b2caps.get(b'delta-compression', ())
        return set(c for c in cap_set if c in _NAME_TO_REVLOG_COMP)

    @classmethod
    def from_revlog_compression(
        cls,
        revlog_encoding: i_comp.RevlogCompHeader,
    ) -> WireDeltaCompression:
        if revlog_encoding is None:
            return cls.NO_COMPRESSION
        elif revlog_encoding == i_comp.REVLOG_COMP_NONE:
            return cls.UNCOMPRESSED
        elif revlog_encoding == b'u':
            return cls.UNCOMPRESSED
        elif revlog_encoding == i_comp.REVLOG_COMP_ZLIB:
            return cls.ZLIB_COMPRESSED
        elif revlog_encoding == i_comp.REVLOG_COMP_ZSTD:
            return cls.ZSTD_COMPRESSED
        else:
            raise ValueError("unknown revlog compression: %s", revlog_encoding)

    def to_revlog_compression(self) -> i_comp.RevlogCompHeader | None:
        if self == self.NO_COMPRESSION:
            return None
        elif self == self.UNCOMPRESSED:
            return i_comp.REVLOG_COMP_NONE
        elif self == self.ZLIB_COMPRESSED:
            return i_comp.REVLOG_COMP_ZLIB
        elif self == self.ZSTD_COMPRESSED:
            return i_comp.REVLOG_COMP_ZSTD
        else:
            assert False, "forgot to implement conversion for %s" % self


@attr.s(slots=True)
class _DeltaHeader:
    """internal data class to hold information read from the binary header

    Hold default values to be used by older format that don't carry that
    information.
    """

    node = attr.ib(type=NodeIdT)
    p1 = attr.ib(type=NodeIdT)
    p2 = attr.ib(type=NodeIdT)
    delta_base = attr.ib(type=NodeIdT)
    cs = attr.ib(type=NodeIdT)
    flags = attr.ib(type=int, default=0)
    protocol_flags = attr.ib(type=int, default=0)
    snapshot_level = attr.ib(type=Optional[int], default=None)
    raw_text_size = attr.ib(type=Optional[int], default=None)
    compression = attr.ib(type=Optional[i_comp.RevlogCompHeader], default=None)
    quality = attr.ib(type=Optional[i_repo.IDeltaQuality], default=None)
    storage_delta_base = attr.ib(type=Optional[NodeIdT], default=None)
    storage_snapshot_level = attr.ib(type=Optional[int], default=None)


class cg1unpacker(i_cg.IChangeGroupUnpacker):
    """Unpacker for cg1 changegroup streams.

    A changegroup unpacker handles the framing of the revision data in
    the wire format. Most consumers will want to use the apply()
    method to add the changes from the changegroup to a repository.

    If you're forwarding a changegroup unmodified to another consumer,
    use getchunks(), which returns an iterator of changegroup
    chunks. This is mostly useful for cases where you need to know the
    data stream has ended by observing the end of the changegroup.

    deltachunk() is useful only if you're applying delta data. Most
    consumers should prefer apply() instead.

    A few other public methods exist. Those are used only for
    bundlerepo and some debug commands - their use is discouraged.
    """

    deltaheader = _CHANGEGROUPV1_DELTA_HEADER
    deltaheadersize = deltaheader.size
    version = b'01'
    _grouplistcount = 1  # One list of files after the manifests

    has_censor_flag = False
    has_filelog_hasmeta_flag = False

    def __init__(self, fh, alg, extras=None):
        if alg is None:
            alg = b'UN'
        if alg not in util.compengines.supportedbundletypes:
            raise error.Abort(_(b'unknown stream compression type: %s') % alg)
        if alg == b'BZ':
            alg = b'_truncatedBZ'

        compengine = util.compengines.forbundletype(alg)
        # Assumes that the result has all io.BytesIO method even when it has
        # not. This work around the fact the typing does not distinct pure
        # stream processing (most common case) and the seekable/closable case,
        # used for bundlerepo & Co.
        #
        # This is not ideal and should be improved at some point.
        stream = compengine.decompressorreader(fh)
        assert stream is not None
        self._stream: io.BytesIO = cast(io.BytesIO, stream)
        self._type = alg
        self.extras = extras or {}
        self.callback = None

    # These methods (compressed, read, seek, tell) all appear to only
    # be used by bundlerepo, but it's a little hard to tell.
    def compressed(self):
        return self._type is not None and self._type != b'UN'

    def read(self, l):
        return self._stream.read(l)

    def seek(self, pos):
        return self._stream.seek(pos)

    def tell(self):
        return self._stream.tell()

    def close(self):
        return self._stream.close()

    def _chunklength(self):
        d = readexactly(self._stream, 4)
        l = struct.unpack(b">l", d)[0]
        if l <= 4:
            if l:
                raise error.Abort(_(b"invalid chunk length %d") % l)
            return 0
        if self.callback:
            self.callback()
        return l - 4

    def changelogheader(self):
        """v10 does not have a changelog header chunk"""
        return {}

    def manifestheader(self):
        """v10 does not have a manifest header chunk"""
        return {}

    def filelogheader(self):
        """return the header of the filelogs chunk, v10 only has the filename"""
        l = self._chunklength()
        if not l:
            return {}
        fname = readexactly(self._stream, l)
        return {b'filename': fname}

    def _deltaheader(
        self,
        headertuple: tuple,
        prevnode: NodeIdT,
    ) -> _DeltaHeader:
        node, p1, p2, cs = headertuple
        if prevnode is None:
            deltabase = p1
        else:
            deltabase = prevnode
        return _DeltaHeader(
            node=node,
            p1=p1,
            p2=p2,
            delta_base=deltabase,
            cs=cs,
        )

    def deltachunk(
        self, prevnode: bytes | None
    ) -> revlogutils.InboundRevision | None:
        l = self._chunklength()
        if not l:
            return None
        headerdata = readexactly(self._stream, self.deltaheadersize)
        header = self.deltaheader.unpack(headerdata)
        chunk = readexactly(self._stream, l - self.deltaheadersize)
        header = self._deltaheader(header, prevnode)
        protocol_flags = header.protocol_flags
        if protocol_flags & storageutil.CG_FLAG_FULL_TEXT:
            protocol_flags &= ~storageutil.CG_FLAG_FULL_TEXT
            delta = None
            text = chunk
        else:
            delta = chunk
            text = None
        return revlogutils.InboundRevision(
            header.node,
            header.p1,
            header.p2,
            header.cs,
            header.delta_base,
            delta,
            header.flags,
            {},
            protocol_flags,
            snapshot_level=header.snapshot_level,
            raw_text_size=header.raw_text_size,
            raw_text=text,
            has_censor_flag=self.has_censor_flag,
            has_filelog_hasmeta_flag=self.has_filelog_hasmeta_flag,
            compression=header.compression,
            other_storage_delta_base=header.storage_delta_base,
            other_storage_snapshot_level=header.storage_snapshot_level,
        )

    def getchunks(self):
        """returns all the chunks contains in the bundle

        Used when you need to forward the binary stream to a file or another
        network API. To do so, it parse the changegroup data, otherwise it will
        block in case of sshrepo because it don't know the end of the stream.
        """
        # For changegroup 1 and 2, we expect 3 parts: changelog, manifestlog,
        # and a list of filelogs. For changegroup 3, we expect 4 parts:
        # changelog, manifestlog, a list of tree manifestlogs, and a list of
        # filelogs.
        #
        # Changelog and manifestlog parts are terminated with empty chunks. The
        # tree and file parts are a list of entry sections. Each entry section
        # is a series of chunks terminating in an empty chunk. The list of these
        # entry sections is terminated in yet another empty chunk, so we know
        # we've reached the end of the tree/file list when we reach an empty
        # chunk that was proceeded by no non-empty chunks.

        parts = 0
        while parts < 2 + self._grouplistcount:
            noentries = True
            while True:
                chunk = getchunk(self)
                if not chunk:
                    # The first two empty chunks represent the end of the
                    # changelog and the manifestlog portions. The remaining
                    # empty chunks represent either A) the end of individual
                    # tree or file entries in the file list, or B) the end of
                    # the entire list. It's the end of the entire list if there
                    # were no entries (i.e. noentries is True).
                    if parts < 2:
                        parts += 1
                    elif noentries:
                        parts += 1
                    break
                noentries = False
                yield chunkheader(len(chunk))
                pos = 0
                while pos < len(chunk):
                    next = pos + 2**20
                    yield chunk[pos:next]
                    pos = next
            yield closechunk()

    def _unpackmanifests(
        self,
        repo,
        revmap,
        trp,
        prog,
        addrevisioncb=None,
        debug_info=None,
        delta_base_reuse_policy=None,
    ):
        self.callback = prog.increment
        # no need to check for empty manifest group here:
        # if the result of the merge of 1 and 2 is the same in 3 and 4,
        # no new manifest will be created and the manifest group will
        # be empty during the pull
        self.manifestheader()
        deltas = self.deltaiter()
        storage = repo.manifestlog.getstorage(b'')
        storage.addgroup(
            deltas,
            revmap,
            trp,
            addrevisioncb=addrevisioncb,
            debug_info=debug_info,
            delta_base_reuse_policy=delta_base_reuse_policy,
        )
        prog.complete()
        self.callback = None

    def apply(
        self,
        repo,
        tr,
        srctype,
        url,
        targetphase=phases.draft,
        expectedtotal=None,
        sidedata_categories=None,
        delta_base_reuse_policy=None,
    ):
        """Add the changegroup returned by source.read() to this repo.
        srctype is a string like 'push', 'pull', or 'unbundle'.  url is
        the URL of the repo where this changegroup is coming from.

        Return an integer summarizing the change to this repo:
        - nothing changed or no source: 0
        - more heads than before: 1+added heads (2..n)
        - fewer heads than before: -1-removed heads (-2..-n)
        - number of heads stays the same: 1

        `sidedata_categories` is an optional set of the remote's sidedata wanted
        categories.

        `delta_base_reuse_policy` is an optional argument, when set to a value
        it will control the way the delta contained into the bundle are reused
        when applied in the revlog.

        See `DELTA_BASE_REUSE_*` entry in mercurial.revlogutils.constants.
        """
        repo = repo.unfiltered()

        debug_info = None
        if repo.ui.configbool(b'debug', b'unbundling-stats'):
            debug_info = []

        # Only useful if we're adding sidedata categories. If both peers have
        # the same categories, then we simply don't do anything.
        adding_sidedata = (
            (
                requirements.REVLOGV2_REQUIREMENT in repo.requirements
                or requirements.CHANGELOGV2_REQUIREMENT in repo.requirements
            )
            and self.version == b'05'
            and srctype == b'pull'
        )
        if adding_sidedata:
            sidedata_helpers = sidedatamod.get_sidedata_helpers(
                repo,
                sidedata_categories or set(),
                pull=True,
            )
        else:
            sidedata_helpers = None

        def csmap(x):
            repo.ui.debug(b"add changeset %s\n" % short(x))
            return len(cl)

        def revmap(x):
            return cl.rev(x)

        try:
            # The transaction may already carry source information. In this
            # case we use the top level data. We overwrite the argument
            # because we need to use the top level value (if they exist)
            # in this function.
            srctype = tr.hookargs.setdefault(b'source', srctype)
            tr.hookargs.setdefault(b'url', url)
            repo.hook(
                b'prechangegroup', throw=True, **pycompat.strkwargs(tr.hookargs)
            )

            # write changelog data to temp files so concurrent readers
            # will not see an inconsistent view
            cl = repo.changelog
            cl.delayupdate(tr)
            oldrevcount = len(cl)

            trp = weakref.proxy(tr)
            # pull off the changeset group
            repo.ui.status(_(b"adding changesets\n"))
            clstart = len(cl)
            progress = repo.ui.makeprogress(
                _(b'changesets'), unit=_(b'chunks'), total=expectedtotal
            )
            self.callback = progress.increment

            efilesset = set()
            duprevs = []

            def ondupchangelog(cl, rev):
                if rev < clstart:
                    duprevs.append(rev)  # pytype: disable=attribute-error

            def onchangelog(cl, rev):
                ctx = cl.changelogrevision(rev)
                assert efilesset is not None  # help pytype
                efilesset.update(ctx.files)
                repo.register_changeset(rev, ctx)

            self.changelogheader()
            deltas = self.deltaiter()
            if not cl.addgroup(
                deltas,
                csmap,
                trp,
                alwayscache=True,
                addrevisioncb=onchangelog,
                duplicaterevisioncb=ondupchangelog,
                debug_info=debug_info,
                delta_base_reuse_policy=delta_base_reuse_policy,
            ):
                repo.ui.develwarn(
                    b'applied empty changelog from changegroup',
                    config=b'warn-empty-changegroup',
                )
            efiles = len(efilesset)
            clend = len(cl)
            changesets = clend - clstart
            progress.complete()
            del deltas
            # TODO Python 2.7 removal
            # del efilesset
            efilesset = None
            self.callback = None

            # Keep track of the (non-changelog) revlogs we've updated and their
            # range of new revisions for sidedata rewrite.
            # TODO do something more efficient than keeping the reference to
            # the revlogs, especially memory-wise.
            touched_manifests = {}
            touched_filelogs = {}

            # pull off the manifest group
            repo.ui.status(_(b"adding manifests\n"))
            # We know that we'll never have more manifests than we had
            # changesets.
            progress = repo.ui.makeprogress(
                _(b'manifests'), unit=_(b'chunks'), total=changesets
            )
            on_manifest_rev = None
            if sidedata_helpers:
                if revlog_constants.KIND_MANIFESTLOG in sidedata_helpers[1]:

                    def on_manifest_rev(manifest, rev):
                        range = touched_manifests.get(manifest)
                        if not range:
                            touched_manifests[manifest] = (rev, rev)
                        else:
                            assert rev == range[1] + 1
                            touched_manifests[manifest] = (range[0], rev)

            self._unpackmanifests(
                repo,
                revmap,
                trp,
                progress,
                addrevisioncb=on_manifest_rev,
                debug_info=debug_info,
                delta_base_reuse_policy=delta_base_reuse_policy,
            )

            needfiles = {}
            if repo.ui.configbool(b'server', b'validate'):
                cl = repo.changelog
                ml = repo.manifestlog
                # validate incoming csets have their manifests
                for cset in range(clstart, clend):
                    mfnode = cl.changelogrevision(cset).manifest
                    mfest = ml[mfnode].read_delta_new_entries()
                    # store file nodes we must see
                    for f, n in mfest.items():
                        needfiles.setdefault(f, set()).add(n)

            on_filelog_rev = None
            if sidedata_helpers:
                if revlog_constants.KIND_FILELOG in sidedata_helpers[1]:

                    def on_filelog_rev(filelog, rev):
                        range = touched_filelogs.get(filelog)
                        if not range:
                            touched_filelogs[filelog] = (rev, rev)
                        else:
                            assert rev == range[1] + 1
                            touched_filelogs[filelog] = (range[0], rev)

            # process the files
            repo.ui.status(_(b"adding file changes\n"))
            newrevs, newfiles = _addchangegroupfiles(
                repo,
                self,
                revmap,
                trp,
                efiles,
                needfiles,
                addrevisioncb=on_filelog_rev,
                debug_info=debug_info,
                delta_base_reuse_policy=delta_base_reuse_policy,
                # this is useful to pass it to the addchangegroupfiles
                # explicitly to help the filelog to decide weither or not to
                # pre-process the data.
                use_hasmeta_flag=self.has_filelog_hasmeta_flag,
            )

            if sidedata_helpers:
                if revlog_constants.KIND_CHANGELOG in sidedata_helpers[1]:
                    cl.rewrite_sidedata(
                        trp, sidedata_helpers, clstart, clend - 1
                    )
                for mf, (startrev, endrev) in touched_manifests.items():
                    mf.rewrite_sidedata(trp, sidedata_helpers, startrev, endrev)
                for fl, (startrev, endrev) in touched_filelogs.items():
                    fl.rewrite_sidedata(trp, sidedata_helpers, startrev, endrev)

            # making sure the value exists
            tr.changes.setdefault(b'changegroup-count-changesets', 0)
            tr.changes.setdefault(b'changegroup-count-revisions', 0)
            tr.changes.setdefault(b'changegroup-count-files', 0)
            tr.changes.setdefault(b'changegroup-count-heads', 0)

            # some code use bundle operation for internal purpose. They usually
            # set `ui.quiet` to do this outside of user sight. Size the report
            # of such operation now happens at the end of the transaction, that
            # ui.quiet has not direct effect on the output.
            #
            # To preserve this intend use an inelegant hack, we fail to report
            # the change if `quiet` is set. We should probably move to
            # something better, but this is a good first step to allow the "end
            # of transaction report" to pass tests.
            if not repo.ui.quiet:
                tr.changes[b'changegroup-count-changesets'] += changesets
                tr.changes[b'changegroup-count-revisions'] += newrevs
                tr.changes[b'changegroup-count-files'] += newfiles

            deltaheads = 0
            newrevcount = len(cl)
            heads_removed, heads_added = cl.diffheads(oldrevcount, newrevcount)
            deltaheads += len(heads_added) - len(heads_removed)
            with cl.reading():
                for h in heads_added:
                    if repo[h].closesbranch():
                        deltaheads -= 1

            # see previous comment about checking ui.quiet
            if not repo.ui.quiet:
                tr.changes[b'changegroup-count-heads'] += deltaheads
            repo.invalidatevolatilesets()

            if changesets > 0:
                if b'node' not in tr.hookargs:
                    tr.hookargs[b'node'] = hex(cl.node(clstart))
                    tr.hookargs[b'node_last'] = hex(cl.node(clend - 1))
                    hookargs = dict(tr.hookargs)
                else:
                    hookargs = dict(tr.hookargs)
                    hookargs[b'node'] = hex(cl.node(clstart))
                    hookargs[b'node_last'] = hex(cl.node(clend - 1))
                repo.hook(
                    b'pretxnchangegroup',
                    throw=True,
                    **pycompat.strkwargs(hookargs),
                )

            added = range(clstart, clend)
            phaseall = None
            if srctype in (b'push', b'serve'):
                # Old servers can not push the boundary themselves.
                # New servers won't push the boundary if changeset already
                # exists locally as secret
                #
                # We should not use added here but the list of all change in
                # the bundle
                if repo.publishing():
                    targetphase = phaseall = phases.public
                else:
                    # closer target phase computation

                    # Those changesets have been pushed from the
                    # outside, their phases are going to be pushed
                    # alongside. Therefor `targetphase` is
                    # ignored.
                    targetphase = phaseall = phases.draft
            if added:
                phases.registernew(repo, tr, targetphase, added)
            if phaseall is not None:
                if duprevs:
                    duprevs.extend(added)
                else:
                    duprevs = added
                phases.advanceboundary(repo, tr, phaseall, [], revs=duprevs)
                duprevs = []

            if changesets > 0:

                def runhooks(unused_success):
                    # These hooks run when the lock releases, not when the
                    # transaction closes. So it's possible for the changelog
                    # to have changed since we last saw it.
                    if clstart >= len(repo):
                        return

                    repo.hook(b"changegroup", **pycompat.strkwargs(hookargs))

                    for rev in added:
                        args = hookargs.copy()
                        args[b'node'] = hex(cl.node(rev))
                        del args[b'node_last']
                        repo.hook(b"incoming", **pycompat.strkwargs(args))

                    repo.ui.log(
                        b"incoming",
                        b"%d incoming changes - new heads: %s\n",
                        len(added),
                        b', '.join([hex(c[:6]) for c in heads_added]),
                    )

                tr.addpostclose(
                    b'changegroup-runhooks-%020i' % clstart,
                    lambda tr: repo._afterlock(runhooks),
                )
            if debug_info is not None:
                display_unbundle_debug_info(repo.ui, debug_info)
        finally:
            repo.ui.flush()
        # never return 0 here:
        if deltaheads < 0:
            ret = deltaheads - 1
        else:
            ret = deltaheads + 1
        return ret

    def deltaiter(self) -> Iterator[revlogutils.InboundRevision]:
        """
        returns an iterator of the deltas in this changegroup

        Useful for passing to the underlying storage system to be stored.
        """
        chain = None
        for chunkdata in iter(lambda: self.deltachunk(chain), None):
            yield chunkdata
            chain = chunkdata.node


class cg2unpacker(cg1unpacker):
    """Unpacker for cg2 streams.

    cg2 streams add support for generaldelta, so the delta header
    format is slightly different. All other features about the data
    remain the same.
    """

    deltaheader = _CHANGEGROUPV2_DELTA_HEADER
    deltaheadersize = deltaheader.size
    version = b'02'

    def _deltaheader(
        self,
        headertuple: tuple,
        prevnode: NodeIdT,
    ) -> _DeltaHeader:
        node, p1, p2, deltabase, cs = headertuple
        return _DeltaHeader(
            node=node,
            p1=p1,
            p2=p2,
            delta_base=deltabase,
            cs=cs,
        )


class cg3unpacker(cg2unpacker):
    """Unpacker for cg3 streams.

    cg3 streams add support for exchanging treemanifests and revlog
    flags. It adds the revlog flags to the delta header and an empty chunk
    separating manifests and files.
    """

    deltaheader = _CHANGEGROUPV3_DELTA_HEADER
    deltaheadersize = deltaheader.size
    version = b'03'
    _grouplistcount = 2  # One list of manifests and one list of files

    has_censor_flag = True

    def _deltaheader(
        self,
        headertuple: tuple,
        prevnode: NodeIdT,
    ) -> _DeltaHeader:
        node, p1, p2, deltabase, cs, flags = headertuple
        return _DeltaHeader(
            node=node,
            p1=p1,
            p2=p2,
            delta_base=deltabase,
            cs=cs,
            flags=flags,
        )

    def _unpackmanifests(
        self,
        repo,
        revmap,
        trp,
        prog,
        addrevisioncb=None,
        debug_info=None,
        delta_base_reuse_policy=None,
    ):
        super()._unpackmanifests(
            repo,
            revmap,
            trp,
            prog,
            addrevisioncb=addrevisioncb,
            debug_info=debug_info,
            delta_base_reuse_policy=delta_base_reuse_policy,
        )
        for chunkdata in iter(self.filelogheader, {}):
            # If we get here, there are directory manifests in the changegroup
            d = chunkdata[b"filename"]
            repo.ui.debug(b"adding %s revisions\n" % d)
            deltas = self.deltaiter()
            if not repo.manifestlog.getstorage(d).addgroup(
                deltas,
                revmap,
                trp,
                addrevisioncb=addrevisioncb,
                debug_info=debug_info,
                delta_base_reuse_policy=delta_base_reuse_policy,
            ):
                raise error.Abort(_(b"received dir revlog group is empty"))


class cg4unpacker(cg3unpacker):
    """Changegroup 4 support more advanced flag for each delta.

    - exchange of full-text without wrapping in a delta
    - exchange compressed delta and full text
    - "hasmeta" flag for filelog
    - snapshot level
    - delta quality
    - delta base and snapshot level in the source storage
    """

    deltaheader = _CHANGEGROUPV4_DELTA_HEADER
    deltaheadersize = deltaheader.size
    version = b'04'

    has_filelog_hasmeta_flag = True

    def _deltaheader(
        self,
        headertuple: tuple,
        prevnode: NodeIdT,
    ) -> _DeltaHeader:
        (
            node,
            p1,
            p2,
            deltabase,
            cs,
            flags,
            snapshot_level,
            raw_size,
            encoded_comp,
            protocol_flags,
            stored_base,
            stored_snap_lvl,
        ) = headertuple
        wire_comp = WireDeltaCompression(encoded_comp)
        if snapshot_level < -1:
            snapshot_level = None
        quality = revlogutils.DeltaQuality.from_v1_flags(flags)
        flags &= ~revlog_constants.REVIDX_DELTA_INFO_FLAGS

        return _DeltaHeader(
            node=node,
            p1=p1,
            p2=p2,
            delta_base=deltabase,
            cs=cs,
            flags=flags,
            snapshot_level=snapshot_level,
            raw_text_size=raw_size,
            compression=wire_comp.to_revlog_compression(),
            protocol_flags=protocol_flags,
            quality=quality,
            storage_delta_base=stored_base,
            storage_snapshot_level=stored_snap_lvl,
        )


class cg5unpacker(cg3unpacker):
    """Unpacker for cg5 streams.

    cg5 streams add support for exchanging sidedata.
    """

    deltaheader = _CHANGEGROUPV5_DELTA_HEADER
    deltaheadersize = deltaheader.size
    version = b'05'

    def _deltaheader(
        self,
        headertuple: tuple,
        prevnode: NodeIdT,
    ) -> _DeltaHeader:
        protocol_flags, node, p1, p2, deltabase, cs, flags = headertuple
        return _DeltaHeader(
            node=node,
            p1=p1,
            p2=p2,
            delta_base=deltabase,
            cs=cs,
            flags=flags,
            protocol_flags=protocol_flags,
        )

    def deltachunk(
        self, prevnode: bytes | None
    ) -> revlogutils.InboundRevision | None:
        res = super().deltachunk(prevnode)
        if res is None:
            return res
        assert not res.sidedata

        if res.protocol_flags & storageutil.CG_FLAG_SIDEDATA:
            sidedata_raw = getchunk(self._stream)
            res.sidedata = sidedatamod.deserialize_sidedata(sidedata_raw)
        return res


class headerlessfixup:
    def __init__(self, fh, h):
        self._h = h
        self._fh = fh

    def read(self, n):
        if self._h:
            d, self._h = self._h[:n], self._h[n:]
            if len(d) < n:
                d += readexactly(self._fh, n - len(d))
            return d
        return readexactly(self._fh, n)


def _revisiondeltatochunks(repo, delta, headerfn, full_text_support=False):
    """Serialize a revisiondelta to changegroup chunks."""

    # The captured revision delta may be encoded as a delta against
    # a base revision or as a full revision. The changegroup format
    # requires that everything on the wire be deltas. So for full
    # revisions, we need to invent a header that says to rewrite
    # data.

    if delta.delta is not None:
        prefix, data = b'', delta.delta
    elif delta.revision is not None:
        assert delta.basenode == repo.nullid
        data = delta.revision
        if full_text_support:
            prefix = b''  # a flag will be set to signal it a full text
        else:
            prefix = mdiff.trivialdiffheader(len(data))
    else:
        data = delta.revision
        prefix = mdiff.replacediffheader(delta.baserevisionsize, len(data))

    meta = headerfn(delta)

    yield chunkheader(len(meta) + len(prefix) + len(data))
    yield meta
    if prefix:
        yield prefix
    yield data

    if delta.protocol_flags & storageutil.CG_FLAG_SIDEDATA:
        # Need a separate chunk for sidedata to be able to differentiate
        # "raw delta" length and sidedata length
        sidedata = delta.sidedata
        yield chunkheader(len(sidedata))
        yield sidedata


def _sortnodesellipsis(store, nodes, cl, lookup):
    """Sort nodes for changegroup generation."""
    # Ellipses serving mode.
    #
    # In a perfect world, we'd generate better ellipsis-ified graphs
    # for non-changelog revlogs. In practice, we haven't started doing
    # that yet, so the resulting DAGs for the manifestlog and filelogs
    # are actually full of bogus parentage on all the ellipsis
    # nodes. This has the side effect that, while the contents are
    # correct, the individual DAGs might be completely out of whack in
    # a case like 882681bc3166 and its ancestors (back about 10
    # revisions or so) in the main hg repo.
    #
    # The one invariant we *know* holds is that the new (potentially
    # bogus) DAG shape will be valid if we order the nodes in the
    # order that they're introduced in dramatis personae by the
    # changelog, so what we do is we sort the non-changelog histories
    # by the order in which they are used by the changelog.
    key = lambda n: cl.rev(lookup(n))
    return sorted(nodes, key=key)


def _resolvenarrowrevisioninfo(
    cl,
    store,
    ischangelog,
    rev,
    linkrev,
    linknode,
    clrevtolocalrev,
    fullclnodes,
    precomputedellipsis,
):
    linkparents = precomputedellipsis[linkrev]

    def local(clrev):
        """Turn a changelog revnum into a local revnum.

        The ellipsis dag is stored as revnums on the changelog,
        but when we're producing ellipsis entries for
        non-changelog revlogs, we need to turn those numbers into
        something local. This does that for us, and during the
        changelog sending phase will also expand the stored
        mappings as needed.
        """
        if clrev == nullrev:
            return nullrev

        if ischangelog:
            return clrev

        # Walk the ellipsis-ized changelog breadth-first looking for a
        # change that has been linked from the current revlog.
        #
        # For a flat manifest revlog only a single step should be necessary
        # as all relevant changelog entries are relevant to the flat
        # manifest.
        #
        # For a filelog or tree manifest dirlog however not every changelog
        # entry will have been relevant, so we need to skip some changelog
        # nodes even after ellipsis-izing.
        walk = [clrev]
        while walk:
            p = walk[0]
            walk = walk[1:]
            if p in clrevtolocalrev:
                return clrevtolocalrev[p]
            elif p in fullclnodes:
                walk.extend([pp for pp in cl.parentrevs(p) if pp != nullrev])
            elif p in precomputedellipsis:
                walk.extend(
                    [pp for pp in precomputedellipsis[p] if pp != nullrev]
                )
            else:
                # In this case, we've got an ellipsis with parents
                # outside the current bundle (likely an
                # incremental pull). We "know" that we can use the
                # value of this same revlog at whatever revision
                # is pointed to by linknode. "Know" is in scare
                # quotes because I haven't done enough examination
                # of edge cases to convince myself this is really
                # a fact - it works for all the (admittedly
                # thorough) cases in our testsuite, but I would be
                # somewhat unsurprised to find a case in the wild
                # where this breaks down a bit. That said, I don't
                # know if it would hurt anything.
                for i in range(rev, 0, -1):
                    if store.linkrev(i) == clrev:
                        return i
                # We failed to resolve a parent for this node, so
                # we crash the changegroup construction.
                if hasattr(store, 'target'):
                    target = store.display_id
                else:
                    # some revlog not actually a revlog
                    target = store._revlog.display_id

                raise error.Abort(
                    b"unable to resolve parent while packing '%s' %r"
                    b' for changeset %r' % (target, rev, clrev)
                )

        return nullrev

    if not linkparents or (store.parentrevs(rev) == (nullrev, nullrev)):
        p1, p2 = nullrev, nullrev
    elif len(linkparents) == 1:
        (p1,) = sorted(local(p) for p in linkparents)
        p2 = nullrev
    else:
        p1, p2 = sorted(local(p) for p in linkparents)

    p1node, p2node = store.node(p1), store.node(p2)

    return p1node, p2node, linknode


def deltagroup(
    repo: i_repo.IRepo,
    store: i_repo.IDeltaEmittingStore,
    nodes: list[NodeIdT],
    ischangelog: bool,
    # a node → link node function
    lookup: LookupFnT,
    forcedeltaparentprev: bool,
    topic: bytes | None = None,
    ellipses: bool = False,
    clrevtolocalrev: dict[RevnumT, RevnumT] | None = None,
    fullclnodes: set[NodeIdT] | None = None,
    precomputedellipsis: i_cg.PreComputedEllipsisT | None = None,
    # typing: the sidedata_helpers were a bit too hairy to type when this
    # function was opportunisticly typed. Feel free to fix that.
    sidedata_helpers=None,
    debug_info: dict[str, NeedsTypeHint] | None = None,
    filelog_hasmeta: bool = False,
    accepted_compression: frozenset[i_comp.RevlogCompHeader] = frozenset(),
) -> Iterator[OutboundRevisionT]:
    """Calculate deltas for a set of revisions.

    Is a generator of ``revisiondelta`` instances.

    If topic is not None, progress detail will be generated using this
    topic name (e.g. changesets, manifests, etc).

    See `revlogutil.sidedata.get_sidedata_helpers` for the doc on
    `sidedata_helpers`.

    if filelog_hasmeta is set, rely of the REVIDX_HASMETA flag to indicate metadata.
    """
    if not nodes:
        return

    cl = repo.changelog

    if ischangelog:
        # `hg log` shows changesets in storage order. To preserve order
        # across clones, send out changesets in storage order.
        nodesorder = b'storage'
    elif ellipses:
        nodes = _sortnodesellipsis(store, nodes, cl, lookup)
        nodesorder = b'nodes'
    else:
        nodesorder = None

    # Perform ellipses filtering and revision massaging. We do this before
    # emitrevisions() because a) filtering out revisions creates less work
    # for emitrevisions() b) dropping revisions would break emitrevisions()'s
    # assumptions about delta choices and we would possibly send a delta
    # referencing a missing base revision.
    #
    # Also, calling lookup() has side-effects with regards to populating
    # data structures. If we don't call lookup() for each node or if we call
    # lookup() after the first pass through each node, things can break -
    # possibly intermittently depending on the python hash seed! For that
    # reason, we store a mapping of all linknodes during the initial node
    # pass rather than use lookup() on the output side.
    if ellipses:
        filtered = []
        adjustedparents = {}
        linknodes = {}
        assert clrevtolocalrev is not None

        for node in nodes:
            rev = store.rev(node)
            linknode = lookup(node)
            linkrev = cl.rev(linknode)
            clrevtolocalrev[linkrev] = rev

            # If linknode is in fullclnodes, it means the corresponding
            # changeset was a full changeset and is being sent unaltered.
            if linknode in fullclnodes:
                linknodes[node] = linknode

            # If the corresponding changeset wasn't in the set computed
            # as relevant to us, it should be dropped outright.
            elif linkrev not in precomputedellipsis:
                continue

            else:
                # We could probably do this later and avoid the dict
                # holding state. But it likely doesn't matter.
                p1node, p2node, linknode = _resolvenarrowrevisioninfo(
                    cl,
                    store,
                    ischangelog,
                    rev,
                    linkrev,
                    linknode,
                    clrevtolocalrev,
                    fullclnodes,
                    precomputedellipsis,
                )

                adjustedparents[node] = (p1node, p2node)
                linknodes[node] = linknode

            filtered.append(node)

        nodes = filtered

    # We expect the first pass to be fast, so we only engage the progress
    # meter for constructing the revision deltas.
    progress = None
    if topic is not None:
        progress = repo.ui.makeprogress(
            topic, unit=_(b'chunks'), total=len(nodes)
        )

    configtarget = repo.ui.config(b'devel', b'bundle.delta')
    if configtarget not in (b'', b'p1', b'full', b'prev'):
        msg = _(b"""config "devel.bundle.delta" as unknown value: %s""")
        repo.ui.warn(msg % configtarget)

    deltamode = i_repo.CG_DELTAMODE_STD
    if forcedeltaparentprev or configtarget == b'prev':
        deltamode = i_repo.CG_DELTAMODE_PREV
    elif configtarget == b'p1':
        deltamode = i_repo.CG_DELTAMODE_P1
    elif configtarget == b'full':
        deltamode = i_repo.CG_DELTAMODE_FULL

    kwargs = {}
    if filelog_hasmeta:
        kwargs["use_hasmeta_flag"] = True

    revisions = store.emitrevisions(
        nodes,
        nodesorder=nodesorder,
        revisiondata=True,
        assumehaveparentrevisions=not ellipses,
        deltamode=deltamode,
        sidedata_helpers=sidedata_helpers,
        debug_info=debug_info,
        accepted_compression=accepted_compression,
        **kwargs,
    )

    for i, revision in enumerate(revisions):
        if progress:
            progress.update(i + 1)

        if ellipses:
            linknode = linknodes[revision.node]

            if revision.node in adjustedparents:
                p1node, p2node = adjustedparents[revision.node]
                revision.p1node = p1node
                revision.p2node = p2node
                revision.flags |= i_repo.REVISION_FLAG_ELLIPSIS

        else:
            linknode = lookup(revision.node)

        revision.linknode = linknode
        yield revision

    if progress:
        progress.complete()


def make_debug_info():
    """ "build a "new" debug_info dictionnary

    That dictionnary can be used to gather information about the bundle process
    """
    return {
        'revision-total': 0,
        'revision-changelog': 0,
        'revision-manifest': 0,
        'revision-files': 0,
        'file-count': 0,
        'merge-total': 0,
        'available-delta': 0,
        'available-full': 0,
        'delta-against-prev': 0,
        'delta-full': 0,
        'delta-against-p1': 0,
        'denied-delta-candeltafn': 0,
        'denied-base-not-available': 0,
        'reused-storage-delta': 0,
        'computed-delta': 0,
    }


def merge_debug_info(base, other):
    """merge the debug information from <other> into <base>

    This function can be used to gather lower level information into higher level ones.
    """
    for key in (
        'revision-total',
        'revision-changelog',
        'revision-manifest',
        'revision-files',
        'merge-total',
        'available-delta',
        'available-full',
        'delta-against-prev',
        'delta-full',
        'delta-against-p1',
        'denied-delta-candeltafn',
        'denied-base-not-available',
        'reused-storage-delta',
        'computed-delta',
    ):
        base[key] += other[key]


_KEY_PART_WIDTH = 17


def _dbg_bdl_line(
    ui,
    indent,
    key,
    base_value=None,
    percentage_base=None,
    percentage_key=None,
    percentage_ref=None,
    extra=None,
):
    """Print one line of debug_bundle_debug_info"""
    line = b"DEBUG-BUNDLING: "
    line += b' ' * (2 * indent)
    key += b":"
    if base_value is not None:
        assert len(key) + 1 + (2 * indent) <= _KEY_PART_WIDTH
        line += key.ljust(_KEY_PART_WIDTH - (2 * indent))
        line += b"%10d" % base_value
    else:
        line += key

    if percentage_base is not None:
        assert base_value is not None
        percentage = base_value * 100 // percentage_base
        if percentage_key is not None:
            line += b" (%d%% of %s %d)" % (
                percentage,
                percentage_key,
                percentage_ref,
            )
        else:
            line += b" (%d%%)" % percentage

    if extra:
        line += b" "
        line += extra

    line += b'\n'
    ui.write_err(line)


def display_bundling_debug_info(
    ui,
    debug_info,
    cl_debug_info,
    mn_debug_info,
    fl_debug_info,
):
    """display debug information gathered during a bundling through `ui`"""
    d = debug_info
    c = cl_debug_info
    m = mn_debug_info
    f = fl_debug_info
    all_info = [
        (b"changelog", b"cl", c),
        (b"manifests", b"mn", m),
        (b"files", b"fl", f),
    ]
    _dbg_bdl_line(ui, 0, b'revisions', d['revision-total'])
    _dbg_bdl_line(ui, 1, b'changelog', d['revision-changelog'])
    _dbg_bdl_line(ui, 1, b'manifest', d['revision-manifest'])
    extra = b'(for %d revlogs)' % d['file-count']
    _dbg_bdl_line(ui, 1, b'files', d['revision-files'], extra=extra)
    if d['merge-total']:
        _dbg_bdl_line(ui, 1, b'merge', d['merge-total'], d['revision-total'])
    for k, __, v in all_info:
        if v['merge-total']:
            _dbg_bdl_line(ui, 2, k, v['merge-total'], v['revision-total'])

    _dbg_bdl_line(ui, 0, b'deltas')
    _dbg_bdl_line(
        ui,
        1,
        b'from-storage',
        d['reused-storage-delta'],
        percentage_base=d['available-delta'],
        percentage_key=b"available",
        percentage_ref=d['available-delta'],
    )

    if d['denied-delta-candeltafn']:
        _dbg_bdl_line(ui, 2, b'denied-fn', d['denied-delta-candeltafn'])
    for __, k, v in all_info:
        if v['denied-delta-candeltafn']:
            _dbg_bdl_line(ui, 3, k, v['denied-delta-candeltafn'])

    if d['denied-base-not-available']:
        _dbg_bdl_line(ui, 2, b'denied-nb', d['denied-base-not-available'])
    for k, __, v in all_info:
        if v['denied-base-not-available']:
            _dbg_bdl_line(ui, 3, k, v['denied-base-not-available'])

    if d['computed-delta']:
        _dbg_bdl_line(ui, 1, b'computed', d['computed-delta'])

    if d['available-full']:
        _dbg_bdl_line(
            ui,
            2,
            b'full',
            d['delta-full'],
            percentage_base=d['available-full'],
            percentage_key=b"native",
            percentage_ref=d['available-full'],
        )
    for k, __, v in all_info:
        if v['available-full']:
            _dbg_bdl_line(
                ui,
                3,
                k,
                v['delta-full'],
                percentage_base=v['available-full'],
                percentage_key=b"native",
                percentage_ref=v['available-full'],
            )

    if d['delta-against-prev']:
        _dbg_bdl_line(ui, 2, b'previous', d['delta-against-prev'])
    for k, __, v in all_info:
        if v['delta-against-prev']:
            _dbg_bdl_line(ui, 3, k, v['delta-against-prev'])

    if d['delta-against-p1']:
        _dbg_bdl_line(ui, 2, b'parent-1', d['delta-against-prev'])
    for k, __, v in all_info:
        if v['delta-against-p1']:
            _dbg_bdl_line(ui, 3, k, v['delta-against-p1'])


class cgpacker(i_cg.IChangeGroupPacker):
    _send_full_text: bool = False
    """Can we send full text without wrapping them in a patch ?"""

    _oldmatcher: MatcherT
    _matcher: MatcherT

    version: bytes
    _forcedeltaparentprev: bool
    _manifestsend: bytes
    _ellipses: bool
    _filelog_hasmeta: bool

    # Set of capabilities we can use to build the bundle.
    _bundlecaps: set
    _remote_sidedata: set
    _isshallow: bool
    _fullclnodes: set[NodeIdT]
    _accepted_compression: frozenset[i_comp.RevlogCompHeader]

    _precomputedellipsis: i_cg.PreComputedEllipsisT | None

    _repo: RepoT

    _verbosenote: Callable[[bytes], None]

    def _init(
        self,
        repo: RepoT,
        oldmatcher: MatcherT,
        matcher: MatcherT,
        version: bytes,
        manifestsend: bytes,
        forcedeltaparentprev: bool = False,
        bundlecaps: set | None = None,
        ellipses: bool = False,
        shallow: bool = False,
        ellipsisroots: i_cg.PreComputedEllipsisT | None = None,
        fullnodes: set[NodeIdT] | None = None,
        remote_sidedata: set[bytes] | None = None,
        filelog_hasmeta: bool = False,
        accepted_compression: frozenset[i_comp.RevlogCompHeader] = frozenset(),
    ):
        """Given a source repo, construct a bundler.

        oldmatcher is a matcher that matches on files the client already has.
        These will not be included in the changegroup.

        matcher is a matcher that matches on files to include in the
        changegroup. Used to facilitate sparse changegroups.

        forcedeltaparentprev indicates whether delta parents must be against
        the previous revision in a delta group. This should only be used for
        compatibility with changegroup version 1.

        builddeltaheader is a callable that constructs the header for a group
        delta.

        manifestsend is a chunk to send after manifests have been fully emitted.

        ellipses indicates whether ellipsis serving mode is enabled.

        bundlecaps is optional and can be used to specify the set of
        capabilities which can be used to build the bundle. While bundlecaps is
        unused in core Mercurial, extensions rely on this feature to communicate
        capabilities to customize the changegroup packer.

        shallow indicates whether shallow data might be sent. The packer may
        need to pack file contents not introduced by the changes being packed.

        fullnodes is the set of changelog nodes which should not be ellipsis
        nodes. We store this rather than the set of nodes that should be
        ellipsis because for very large histories we expect this to be
        significantly smaller.

        remote_sidedata is the set of sidedata categories wanted by the remote.

        filelog_hasmeta: when true, set the has_meta flag when appropriate
        (instead of relying on parent order).
        """
        assert oldmatcher
        assert matcher
        self._oldmatcher = oldmatcher
        self._matcher = matcher

        self.version = version
        self._forcedeltaparentprev = forcedeltaparentprev
        self._manifestsend = manifestsend
        self._ellipses = ellipses
        self._filelog_hasmeta = filelog_hasmeta
        self._accepted_compression = accepted_compression

        # Set of capabilities we can use to build the bundle.
        if bundlecaps is None:
            bundlecaps = set()
        self._bundlecaps = bundlecaps
        if remote_sidedata is None:
            remote_sidedata = set()
        self._remote_sidedata = remote_sidedata
        self._isshallow = shallow
        if fullnodes is None:
            fullnodes = set()
        self._fullclnodes = fullnodes

        # Maps ellipsis revs to their roots at the changelog level.
        self._precomputedellipsis = ellipsisroots

        self._repo = repo

        if self._repo.ui.verbose and not self._repo.ui.debugflag:
            self._verbosenote = self._repo.ui.note
        else:
            self._verbosenote = lambda s: None

    @abc.abstractmethod
    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        ...

    def generate(
        self,
        commonrevs: set[RevnumT],
        clnodes: list[NodeIdT],
        fastpathlinkrev: bool,
        source: bytes,
        changelog: bool = True,
    ) -> Iterator[bytes]:
        """Yield a sequence of changegroup byte chunks.
        If changelog is False, changelog data won't be added to changegroup
        """

        debug_info = None
        repo = self._repo
        if repo.ui.configbool(b'debug', b'bundling-stats'):
            debug_info = make_debug_info()
        cl = repo.changelog

        self._verbosenote(_(b'uncompressed size of bundle content:\n'))
        size = 0

        sidedata_helpers = None
        if self.version == b'05':
            remote_sidedata = self._remote_sidedata
            if source == b'strip':
                # We're our own remote when stripping, get the no-op helpers
                # TODO a better approach would be for the strip bundle to
                # correctly advertise its sidedata categories directly.
                remote_sidedata = repo._wanted_sidedata
            sidedata_helpers = sidedatamod.get_sidedata_helpers(
                repo,
                remote_sidedata,
            )

        cl_debug_info = None
        if debug_info is not None:
            cl_debug_info = make_debug_info()
        clstate, deltas = self._generatechangelog(
            cl,
            clnodes,
            generate=changelog,
            sidedata_helpers=sidedata_helpers,
            debug_info=cl_debug_info,
        )
        for delta in deltas:
            for chunk in _revisiondeltatochunks(
                self._repo,
                delta,
                self._builddeltaheader,
                full_text_support=self._send_full_text,
            ):
                size += len(chunk)
                yield chunk

        close = closechunk()
        size += len(close)
        yield closechunk()
        if debug_info is not None:
            merge_debug_info(debug_info, cl_debug_info)
            debug_info['revision-changelog'] = cl_debug_info['revision-total']

        self._verbosenote(_(b'%8.i (changelog)\n') % size)

        clrevorder = clstate[b'clrevorder']
        manifests = clstate[b'manifests']
        changedfiles = clstate[b'changedfiles']

        if debug_info is not None:
            debug_info['file-count'] = len(changedfiles)

        # We need to make sure that the linkrev in the changegroup refers to
        # the first changeset that introduced the manifest or file revision.
        # The fastpath is usually safer than the slowpath, because the filelogs
        # are walked in revlog order.
        #
        # When taking the slowpath when the manifest revlog uses generaldelta,
        # the manifest may be walked in the "wrong" order. Without 'clrevorder',
        # we would get an incorrect linkrev (see fix in cc0ff93d0c0c).
        #
        # When taking the fastpath, we are only vulnerable to reordering
        # of the changelog itself. The changelog never uses generaldelta and is
        # never reordered. To handle this case, we simply take the slowpath,
        # which already has the 'clrevorder' logic. This was also fixed in
        # cc0ff93d0c0c.

        # Treemanifests don't work correctly with fastpathlinkrev
        # either, because we don't discover which directory nodes to
        # send along with files. This could probably be fixed.
        fastpathlinkrev = fastpathlinkrev and not scmutil.istreemanifest(repo)

        fnodes = {}  # needed file nodes

        size = 0
        mn_debug_info = None
        if debug_info is not None:
            mn_debug_info = make_debug_info()
        it = self.generatemanifests(
            commonrevs,
            clrevorder,
            fastpathlinkrev,
            manifests,
            fnodes,
            source,
            clstate[b'clrevtomanifestrev'],
            sidedata_helpers=sidedata_helpers,
            debug_info=mn_debug_info,
        )

        for tree, deltas in it:
            if tree:
                assert self.version in (b'03', b'05')
                chunk = _fileheader(tree)
                size += len(chunk)
                yield chunk

            for delta in deltas:
                chunks = _revisiondeltatochunks(
                    self._repo,
                    delta,
                    self._builddeltaheader,
                    full_text_support=self._send_full_text,
                )
                for chunk in chunks:
                    size += len(chunk)
                    yield chunk

            close = closechunk()
            size += len(close)
            yield close
        if debug_info is not None:
            merge_debug_info(debug_info, mn_debug_info)
            debug_info['revision-manifest'] = mn_debug_info['revision-total']

        self._verbosenote(_(b'%8.i (manifests)\n') % size)
        yield self._manifestsend

        mfdicts = None
        if self._ellipses and self._isshallow:
            mfdicts = [
                (repo.manifestlog[n].read(), lr)
                for (n, lr) in pycompat.iteritems(manifests)
            ]

        manifests.clear()
        clrevs = {cl.rev(x) for x in clnodes}

        fl_debug_info = None
        if debug_info is not None:
            fl_debug_info = make_debug_info()
        it = self.generatefiles(
            changedfiles,
            commonrevs,
            source,
            mfdicts,
            fastpathlinkrev,
            fnodes,
            clrevs,
            sidedata_helpers=sidedata_helpers,
            debug_info=fl_debug_info,
        )

        for path, deltas in it:
            h = _fileheader(path)
            size = len(h)
            yield h

            for delta in deltas:
                chunks = _revisiondeltatochunks(
                    self._repo,
                    delta,
                    self._builddeltaheader,
                    full_text_support=self._send_full_text,
                )
                for chunk in chunks:
                    size += len(chunk)
                    yield chunk

            close = closechunk()
            size += len(close)
            yield close

            self._verbosenote(_(b'%8.i  %s\n') % (size, path))

        yield closechunk()
        if debug_info is not None:
            merge_debug_info(debug_info, fl_debug_info)
            debug_info['revision-files'] = fl_debug_info['revision-total']

        if debug_info is not None:
            display_bundling_debug_info(
                repo.ui,
                debug_info,
                cl_debug_info,
                mn_debug_info,
                fl_debug_info,
            )

        if clnodes:
            repo.hook(b'outgoing', node=hex(clnodes[0]), source=source)

    def _generatechangelog(
        self,
        cl,
        nodes,
        generate=True,
        sidedata_helpers=None,
        debug_info=None,
    ) -> tuple[dict[bytes, NeedsTypeHint], Iterator[OutboundRevisionT]]:
        """Generate data for changelog chunks.

        Returns a 2-tuple of a dict containing state and an iterable of
        byte chunks. The state will not be fully populated until the
        chunk stream has been fully consumed.

        if generate is False, the state will be fully populated and no chunk
        stream will be yielded

        See `revlogutil.sidedata.get_sidedata_helpers` for the doc on
        `sidedata_helpers`.
        """
        clrevorder = {}
        manifests = {}
        mfl = self._repo.manifestlog
        changedfiles = set()
        clrevtomanifestrev = {}

        state = {
            b'clrevorder': clrevorder,
            b'manifests': manifests,
            b'changedfiles': changedfiles,
            b'clrevtomanifestrev': clrevtomanifestrev,
        }

        if not (generate or self._ellipses):
            # sort the nodes in storage order
            nodes = sorted(nodes, key=cl.rev)
            for node in nodes:
                c = cl.changelogrevision(node)
                clrevorder[node] = len(clrevorder)
                # record the first changeset introducing this manifest version
                manifests.setdefault(c.manifest, node)
                # Record a complete list of potentially-changed files in
                # this manifest.
                changedfiles.update(c.files)

            return state, iter(())

        # Callback for the changelog, used to collect changed files and
        # manifest nodes.
        # Returns the linkrev node (identity in the changelog case).
        def lookupcl(x):
            c = cl.changelogrevision(x)
            clrevorder[x] = len(clrevorder)

            if self._ellipses:
                # Only update manifests if x is going to be sent. Otherwise we
                # end up with bogus linkrevs specified for manifests and
                # we skip some manifest nodes that we should otherwise
                # have sent.
                if (
                    x in self._fullclnodes
                    or cl.rev(x) in self._precomputedellipsis
                ):
                    manifestnode = c.manifest
                    # Record the first changeset introducing this manifest
                    # version.
                    manifests.setdefault(manifestnode, x)
                    # Set this narrow-specific dict so we have the lowest
                    # manifest revnum to look up for this cl revnum. (Part of
                    # mapping changelog ellipsis parents to manifest ellipsis
                    # parents)
                    clrevtomanifestrev.setdefault(
                        cl.rev(x), mfl.rev(manifestnode)
                    )
                # We can't trust the changed files list in the changeset if the
                # client requested a shallow clone.
                if self._isshallow:
                    changedfiles.update(mfl[c.manifest].read().keys())
                else:
                    changedfiles.update(c.files)
            else:
                # record the first changeset introducing this manifest version
                manifests.setdefault(c.manifest, x)
                # Record a complete list of potentially-changed files in
                # this manifest.
                changedfiles.update(c.files)

            return x

        gen = deltagroup(
            self._repo,
            cl,
            nodes,
            True,
            lookupcl,
            self._forcedeltaparentprev,
            ellipses=self._ellipses,
            topic=_(b'changesets'),
            clrevtolocalrev={},
            fullclnodes=self._fullclnodes,
            precomputedellipsis=self._precomputedellipsis,
            sidedata_helpers=sidedata_helpers,
            debug_info=debug_info,
            accepted_compression=self._accepted_compression,
        )

        return state, gen

    def generatemanifests(
        self,
        commonrevs,
        clrevorder,
        fastpathlinkrev,
        manifests,
        fnodes,
        source,
        clrevtolocalrev,
        sidedata_helpers=None,
        debug_info=None,
    ) -> Iterator[tuple[bytes, Iterator[OutboundRevisionT]]]:
        """Returns an iterator of changegroup chunks containing manifests.

        `source` is unused here, but is used by extensions like remotefilelog to
        change what is sent based in pulls vs pushes, etc.

        See `revlogutil.sidedata.get_sidedata_helpers` for the doc on
        `sidedata_helpers`.
        """
        repo = self._repo
        mfl = repo.manifestlog
        tmfnodes = {b'': manifests}

        # Callback for the manifest, used to collect linkrevs for filelog
        # revisions.
        # Returns the linkrev node (collected in lookupcl).
        def makelookupmflinknode(tree, nodes):
            if fastpathlinkrev:
                assert not tree

                # pytype: disable=unsupported-operands
                return manifests.__getitem__
                # pytype: enable=unsupported-operands

            def lookupmflinknode(x):
                """Callback for looking up the linknode for manifests.

                Returns the linkrev node for the specified manifest.

                SIDE EFFECT:

                1) fclnodes gets populated with the list of relevant
                   file nodes if we're not using fastpathlinkrev
                2) When treemanifests are in use, collects treemanifest nodes
                   to send

                Note that this means manifests must be completely sent to
                the client before you can trust the list of files and
                treemanifests to send.
                """
                clnode = nodes[x]
                mctx = mfl.get(tree, x)
                mdata = mctx.read_delta_parents(shallow=True, exact=False)
                for p, n, fl in mdata.iterentries():
                    if fl == b't':  # subdirectory manifest
                        subtree = tree + p + b'/'
                        tmfclnodes = tmfnodes.setdefault(subtree, {})
                        tmfclnode = tmfclnodes.setdefault(n, clnode)
                        if clrevorder[clnode] < clrevorder[tmfclnode]:
                            tmfclnodes[n] = clnode
                    else:
                        f = tree + p
                        fclnodes = fnodes.setdefault(f, {})
                        fclnode = fclnodes.setdefault(n, clnode)
                        if clrevorder[clnode] < clrevorder[fclnode]:
                            fclnodes[n] = clnode
                return clnode

            return lookupmflinknode

        while tmfnodes:
            tree, nodes = tmfnodes.popitem()

            should_visit = self._matcher.visitdir(tree[:-1])
            if tree and not should_visit:
                continue

            store = mfl.getstorage(tree)

            if not should_visit:
                # No nodes to send because this directory is out of
                # the client's view of the repository (probably
                # because of narrow clones). Do this even for the root
                # directory (tree=='')
                prunednodes = []
            else:
                # Avoid sending any manifest nodes we can prove the
                # client already has by checking linkrevs. See the
                # related comment in generatefiles().
                prunednodes = self._prunemanifests(store, nodes, commonrevs)

            if tree and not prunednodes:
                continue

            lookupfn = makelookupmflinknode(tree, nodes)

            deltas = deltagroup(
                self._repo,
                store,
                prunednodes,
                False,
                lookupfn,
                self._forcedeltaparentprev,
                ellipses=self._ellipses,
                topic=_(b'manifests'),
                clrevtolocalrev=clrevtolocalrev,
                fullclnodes=self._fullclnodes,
                precomputedellipsis=self._precomputedellipsis,
                sidedata_helpers=sidedata_helpers,
                debug_info=debug_info,
                accepted_compression=self._accepted_compression,
            )

            if not self._oldmatcher.visitdir(store.tree[:-1]):
                yield tree, deltas
            else:
                # 'deltas' is a generator and we need to consume it even if
                # we are not going to send it because a side-effect is that
                # it updates tmdnodes (via lookupfn)
                for d in deltas:
                    pass
                if not tree:
                    yield tree, iter([])

    def _prunemanifests(self, store, nodes, commonrevs):
        if not self._ellipses:
            # In non-ellipses case and large repositories, it is better to
            # prevent calling of store.rev and store.linkrev on a lot of
            # nodes as compared to sending some extra data
            return nodes.copy()
        # This is split out as a separate method to allow filtering
        # commonrevs in extension code.
        #
        # TODO(augie): this shouldn't be required, instead we should
        # make filtering of revisions to send delegated to the store
        # layer.
        frev, flr = store.rev, store.linkrev
        return [n for n in nodes if flr(frev(n)) not in commonrevs]

    # The 'source' parameter is useful for extensions
    def generatefiles(
        self,
        changedfiles,
        commonrevs,
        source,
        mfdicts,
        fastpathlinkrev,
        fnodes,
        clrevs,
        sidedata_helpers=None,
        debug_info=None,
    ) -> Iterator[tuple[bytes, Iterator[OutboundRevisionT]]]:
        changedfiles = [
            f
            for f in changedfiles
            if self._matcher(f) and not self._oldmatcher(f)
        ]

        if not fastpathlinkrev:

            def normallinknodes(unused, fname):
                return fnodes.get(fname, {})

        else:
            cln = self._repo.changelog.node

            def normallinknodes(store, fname):
                flinkrev = store.linkrev
                fnode = store.node
                revs = ((r, flinkrev(r)) for r in store)
                return {fnode(r): cln(lr) for r, lr in revs if lr in clrevs}

        clrevtolocalrev = {}

        if self._isshallow:
            # In a shallow clone, the linknodes callback needs to also include
            # those file nodes that are in the manifests we sent but weren't
            # introduced by those manifests.
            commonctxs = [self._repo[c] for c in commonrevs]
            clrev = self._repo.changelog.rev

            def linknodes(flog, fname):
                for c in commonctxs:
                    try:
                        fnode = c.filenode(fname)
                        clrevtolocalrev[c.rev()] = flog.rev(fnode)
                    except error.ManifestLookupError:
                        pass
                links = normallinknodes(flog, fname)
                if len(links) != len(mfdicts):
                    for mf, lr in mfdicts:
                        fnode = mf.get(fname, None)
                        if fnode in links:
                            links[fnode] = min(links[fnode], lr, key=clrev)
                        elif fnode:
                            links[fnode] = lr
                return links

        else:
            linknodes = normallinknodes

        repo = self._repo
        progress = repo.ui.makeprogress(
            _(b'files'), unit=_(b'files'), total=len(changedfiles)
        )
        for i, fname in enumerate(sorted(changedfiles)):
            filerevlog = repo.file(fname)
            if not filerevlog:
                raise error.Abort(
                    _(b"empty or missing file data for %s") % fname
                )

            clrevtolocalrev.clear()

            linkrevnodes = linknodes(filerevlog, fname)

            # Lookup for filenodes, we collected the linkrev nodes above in the
            # fastpath case and with lookupmf in the slowpath case.
            def lookupfilelog(x):
                return linkrevnodes[x]

            frev, flr = filerevlog.rev, filerevlog.linkrev
            # Skip sending any filenode we know the client already
            # has. This avoids over-sending files relatively
            # inexpensively, so it's not a problem if we under-filter
            # here.
            filenodes = [
                n for n in linkrevnodes if flr(frev(n)) not in commonrevs
            ]

            if not filenodes:
                continue

            progress.update(i + 1, item=fname)

            deltas = deltagroup(
                self._repo,
                filerevlog,
                filenodes,
                False,
                lookupfilelog,
                self._forcedeltaparentprev,
                ellipses=self._ellipses,
                clrevtolocalrev=clrevtolocalrev,
                fullclnodes=self._fullclnodes,
                precomputedellipsis=self._precomputedellipsis,
                sidedata_helpers=sidedata_helpers,
                debug_info=debug_info,
                filelog_hasmeta=self._filelog_hasmeta,
                accepted_compression=self._accepted_compression,
            )

            yield fname, deltas

        progress.complete()


class ChangeGroupPacker01(cgpacker):
    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        return _CHANGEGROUPV1_DELTA_HEADER.pack(
            d.node,
            d.p1node,
            d.p2node,
            d.linknode,
        )

    def __init__(
        self,
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=False,
        shallow=False,
        ellipsisroots=None,
        fullnodes=None,
        remote_sidedata=None,
    ):
        super()._init(
            repo,
            oldmatcher,
            matcher,
            b'01',
            manifestsend=b'',
            forcedeltaparentprev=True,
            bundlecaps=bundlecaps,
            ellipses=ellipses,
            shallow=shallow,
            ellipsisroots=ellipsisroots,
            fullnodes=fullnodes,
        )


class ChangeGroupPacker02(cgpacker):
    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        return _CHANGEGROUPV2_DELTA_HEADER.pack(
            d.node,
            d.p1node,
            d.p2node,
            d.basenode,
            d.linknode,
        )

    def __init__(
        self,
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=False,
        shallow=False,
        ellipsisroots=None,
        fullnodes=None,
        remote_sidedata=None,
    ):
        super()._init(
            repo,
            oldmatcher,
            matcher,
            b'02',
            manifestsend=b'',
            bundlecaps=bundlecaps,
            ellipses=ellipses,
            shallow=shallow,
            ellipsisroots=ellipsisroots,
            fullnodes=fullnodes,
        )


class ChangeGroupPacker03(cgpacker):
    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        return _CHANGEGROUPV3_DELTA_HEADER.pack(
            d.node,
            d.p1node,
            d.p2node,
            d.basenode,
            d.linknode,
            d.flags,
        )

    def __init__(
        self,
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=False,
        shallow=False,
        ellipsisroots=None,
        fullnodes=None,
        remote_sidedata=None,
    ):
        super()._init(
            repo,
            oldmatcher,
            matcher,
            b'03',
            manifestsend=closechunk(),
            bundlecaps=bundlecaps,
            ellipses=ellipses,
            shallow=shallow,
            ellipsisroots=ellipsisroots,
            fullnodes=fullnodes,
        )


class ChangeGroupPacker04(cgpacker):
    """Changegroup 4 support more advanced flag for each delta.

    see documentation of cg4unpacker for details.
    """

    _send_full_text = True

    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        snap_lvl = -2  # means no-info
        if d.snapshot_level is not None:
            snap_lvl = d.snapshot_level
        wire_comp = WireDeltaCompression.from_revlog_compression(d.compression)

        protocol_flags = 0
        if d.delta is None and d.revision is not None:
            protocol_flags |= storageutil.CG_FLAG_FULL_TEXT
        flags = d.flags
        if d.quality is not None:
            flags |= d.quality.to_v1_flags()
        return _CHANGEGROUPV4_DELTA_HEADER.pack(
            d.node,
            d.p1node,
            d.p2node,
            d.basenode,
            d.linknode,
            flags,
            snap_lvl,
            d.raw_revision_size,
            wire_comp,
            protocol_flags,
            # TODO: When d.basenode == d.storage_delta_base, serializing
            # these is redundant. So we could save space by only included them
            # when they differ having a variable size header.
            #
            # However I suspect that compression may very well deal with this
            # for use, so let duplicate the information and see what happens in
            # practice.
            d.stored_delta_base,
            d.stored_snapshot_level,
        )

    def __init__(
        self,
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=False,
        shallow=False,
        ellipsisroots=None,
        fullnodes=None,
        remote_sidedata=None,
    ):
        super()._init(
            repo,
            oldmatcher,
            matcher,
            b'04',
            manifestsend=closechunk(),
            bundlecaps=bundlecaps,
            ellipses=ellipses,
            shallow=shallow,
            ellipsisroots=ellipsisroots,
            fullnodes=fullnodes,
            filelog_hasmeta=True,
            accepted_compression=WireDeltaCompression.accepted(bundlecaps),
        )


class ChangeGroupPacker05(cgpacker):
    # Sidedata is in a separate chunk from the delta to differentiate
    # "raw delta" and sidedata.
    def _builddeltaheader(self, d: OutboundRevisionT) -> bytes:
        return _CHANGEGROUPV5_DELTA_HEADER.pack(
            d.protocol_flags,
            d.node,
            d.p1node,
            d.p2node,
            d.basenode,
            d.linknode,
            d.flags,
        )

    def __init__(
        self,
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=False,
        shallow=False,
        ellipsisroots=None,
        fullnodes=None,
        remote_sidedata=None,
    ):
        super()._init(
            repo,
            oldmatcher,
            matcher,
            b'05',
            manifestsend=closechunk(),
            bundlecaps=bundlecaps,
            ellipses=ellipses,
            shallow=shallow,
            ellipsisroots=ellipsisroots,
            fullnodes=fullnodes,
            remote_sidedata=remote_sidedata,
        )


_packermap: dict[
    bytes, tuple[type[i_cg.IChangeGroupPacker], type[i_cg.IChangeGroupUnpacker]]
] = {
    b'01': (ChangeGroupPacker01, cg1unpacker),
    # cg2 adds support for exchanging generaldelta
    b'02': (ChangeGroupPacker02, cg2unpacker),
    # cg3 adds support for exchanging revlog flags and treemanifests
    b'03': (ChangeGroupPacker03, cg3unpacker),
    # cg4 adds support for exchanging more advances flags
    b'04': (ChangeGroupPacker04, cg4unpacker),
    # ch5 adds support for exchanging sidedata
    b'05': (ChangeGroupPacker05, cg5unpacker),
}


def allsupportedversions(repo):
    versions = set(_packermap.keys())
    want_v5 = (
        repo.ui.configbool(b'experimental', b'changegroup5')
        or requirements.REVLOGV2_REQUIREMENT in repo.requirements
        or requirements.CHANGELOGV2_REQUIREMENT in repo.requirements
    )
    if not want_v5:
        versions.discard(b'05')
    return versions


# Changegroup versions that can be applied to the repo
def supportedincomingversions(repo):
    versions = allsupportedversions(repo)
    # Using changegroupv4 is most useful when the receiving end support
    # delta-info flags. So we don't advertise support for it by default unless
    # explicitly specified.
    #
    # (It is supported for outgoing bundle unconditionnaly so that we won't
    # need to manually enable it for bundle or pushing it)
    if repo.ui.config_is_set(b'experimental', b'changegroup4'):
        want_v4 = repo.ui.configbool(b'experimental', b'changegroup4')
    else:
        want_v4 = scmutil.use_delta_info(repo)
    if not want_v4:
        versions.discard(b'04')
    return versions


# Changegroup versions that can be created from the repo
def supportedoutgoingversions(repo):
    versions = allsupportedversions(repo)
    if scmutil.istreemanifest(repo):
        # Versions 01 and 02 support only flat manifests and it's just too
        # expensive to convert between the flat manifest and tree manifest on
        # the fly. Since tree manifests are hashed differently, all of history
        # would have to be converted. Instead, we simply don't even pretend to
        # support versions 01 and 02.
        versions.discard(b'01')
        versions.discard(b'02')
    if repo.is_narrow:
        # Versions 01 and 02 don't support revlog flags, and we need to
        # support that for stripping and unbundling to work.
        versions.discard(b'01')
        versions.discard(b'02')
    if LFS_REQUIREMENT in repo.requirements:
        # Versions 01 and 02 don't support revlog flags, and we need to
        # mark LFS entries with REVIDX_EXTSTORED.
        versions.discard(b'01')
        versions.discard(b'02')

    return versions


def localversion(repo):
    """Finds the best version to use for "local bundles"

    "Local bundles"  are meant to be used locally, such as those from strip and
    shelve, and temporary bundles.
    """
    versions = allsupportedversions(repo)
    # only use changegroup4 if explicitly requested or if using a format that
    # directly benefit from it.
    if repo.ui.config_is_set(b'experimental', b'changegroup4'):
        want_v4 = repo.ui.configbool(b'experimental', b'changegroup4')
    else:
        want_v4 = scmutil.use_delta_info(repo)
    if not want_v4:
        versions.discard(b'04')
    return max(versions)


def safeversion(repo):
    # Finds the smallest version that it's safe to assume clients of the repo
    # will support. For example, all hg versions that support generaldelta also
    # support changegroup 02.
    versions = supportedoutgoingversions(repo)
    if requirements.GENERALDELTA_REQUIREMENT in repo.requirements:
        versions.discard(b'01')
    assert versions
    return min(versions)


def getbundler(
    version,
    repo,
    bundlecaps=None,
    oldmatcher=None,
    matcher=None,
    ellipses=False,
    shallow=False,
    ellipsisroots=None,
    fullnodes=None,
    remote_sidedata=None,
) -> i_cg.IChangeGroupPacker:
    assert version in supportedoutgoingversions(repo)

    if matcher is None:
        matcher = matchmod.always()
    if oldmatcher is None:
        oldmatcher = matchmod.never()

    if version == b'01' and not matcher.always():
        raise error.ProgrammingError(
            b'version 01 changegroups do not support sparse file matchers'
        )

    if ellipses and version in (b'01', b'02'):
        raise error.Abort(
            _(
                b'ellipsis nodes require at least cg3 on client and server, '
                b'but negotiated version %s'
            )
            % version
        )

    # Requested files could include files not in the local store. So
    # filter those out.
    matcher = repo.narrowmatch(matcher)

    fn = _packermap[version][0]
    return fn(
        repo,
        oldmatcher,
        matcher,
        bundlecaps,
        ellipses=ellipses,
        shallow=shallow,
        ellipsisroots=ellipsisroots,
        fullnodes=fullnodes,
        remote_sidedata=remote_sidedata,
    )


def getunbundler(
    version: bytes,
    fh: io.BytesIO | NeedsTypeHint,
    alg: bytes | None,
    extras: dict[bytes, NeedsTypeHint] | None = None,
) -> i_cg.IChangeGroupUnpacker:
    return _packermap[version][1](fh, alg, extras=extras)


def _changegroupinfo(repo, nodes, source):
    if repo.ui.verbose or source == b'bundle':
        repo.ui.status(_(b"%d changesets found\n") % len(nodes))
    if repo.ui.debugflag:
        repo.ui.debug(b"list of changesets:\n")
        for node in nodes:
            repo.ui.debug(b"%s\n" % hex(node))


def makechangegroup(
    repo,
    outgoing,
    version,
    source,
    fastpath=False,
    bundlecaps=None,
):
    cgstream = makestream(
        repo,
        outgoing,
        version,
        source,
        fastpath=fastpath,
        bundlecaps=bundlecaps,
    )
    extras = {b'clcount': len(outgoing.missing)}
    # This is fairly basic and signal what compression -might- be used, and not
    # which one is actually used, but this will be good enough for now.
    delta_comp = WireDeltaCompression.filter_known(bundlecaps)
    if delta_comp:
        if requirements.REVLOG_COMPRESSION_ZSTD not in repo.requirements:
            delta_comp.discard(b"zstd")
        if delta_comp:
            extras[b'delta-compression'] = delta_comp
    return getunbundler(
        version,
        util.chunkbuffer(cgstream),
        None,
        extras=extras,
    )


def makestream(
    repo,
    outgoing,
    version,
    source,
    fastpath=False,
    bundlecaps=None,
    matcher=None,
    remote_sidedata=None,
):
    bundler = getbundler(
        version,
        repo,
        bundlecaps=bundlecaps,
        matcher=matcher,
        remote_sidedata=remote_sidedata,
    )

    repo = repo.unfiltered()
    commonrevs = outgoing.common
    csets = outgoing.missing
    heads = outgoing.ancestorsof
    # We go through the fast path if we get told to, or if all (unfiltered
    # heads have been requested (since we then know there all linkrevs will
    # be pulled by the client).
    heads.sort()
    fastpathlinkrev = fastpath or (
        repo.filtername is None and heads == sorted(repo.heads())
    )

    repo.hook(b'preoutgoing', throw=True, source=source)
    _changegroupinfo(repo, csets, source)
    return bundler.generate(
        commonrevs,
        csets,
        fastpathlinkrev,
        source,
    )


def _addchangegroupfiles(
    repo,
    source,
    revmap,
    trp,
    expectedfiles,
    needfiles,
    addrevisioncb=None,
    debug_info=None,
    delta_base_reuse_policy=None,
    use_hasmeta_flag=False,
):
    revisions = 0
    files = 0
    progress = repo.ui.makeprogress(
        _(b'files'), unit=_(b'files'), total=expectedfiles
    )
    for chunkdata in iter(source.filelogheader, {}):
        files += 1
        f = chunkdata[b"filename"]
        repo.ui.debug(b"adding %s revisions\n" % f)
        progress.increment()
        fl = repo.file(f, writable=True)
        o = len(fl)
        try:
            deltas = source.deltaiter()
            added = fl.addgroup(
                deltas,
                revmap,
                trp,
                addrevisioncb=addrevisioncb,
                debug_info=debug_info,
                delta_base_reuse_policy=delta_base_reuse_policy,
                use_hasmeta_flag=use_hasmeta_flag,
            )
            if not added:
                raise error.Abort(_(b"received file revlog group is empty"))
        except error.CensoredBaseError as e:
            raise error.Abort(_(b"received delta base is censored: %s") % e)
        revisions += len(fl) - o
        if f in needfiles:
            needs = needfiles[f]
            for new in range(o, len(fl)):
                n = fl.node(new)
                if n in needs:
                    needs.remove(n)
                else:
                    raise error.Abort(_(b"received spurious file revlog entry"))
            if not needs:
                del needfiles[f]
    progress.complete()

    for f, needs in needfiles.items():
        fl = repo.file(f)
        for n in needs:
            try:
                fl.rev(n)
            except error.LookupError:
                raise error.Abort(
                    _(b'missing file data for %s:%s - run hg verify')
                    % (f, hex(n))
                )

    return revisions, files



# config.py - configuration parsing for Mercurial
#
#  Copyright 2009 Olivia Mackall <olivia@selenic.com> and others
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import errno
import os
import typing

from .i18n import _
from . import (
    encoding,
    error,
    util,
)

if typing.TYPE_CHECKING:
    from .interfaces.types import (
        CfgRemapT,
        CfgSectionsT,
    )


class config:
    def __init__(self, data=None):
        self._current_source_level = 0
        self._data = {}
        self._unset = []
        if data:
            for k in data._data:
                self._data[k] = data[k].copy()
            self._current_source_level = data._current_source_level + 1

    def new_source(self):
        """increment the source counter

        This is used to define source priority when reading"""
        self._current_source_level += 1

    def copy(self):
        return config(self)

    def __contains__(self, section):
        return section in self._data

    def hasitem(self, section, item):
        return item in self._data.get(section, {})

    def __getitem__(self, section):
        return self._data.get(section, {})

    def __iter__(self):
        yield from self.sections()

    def update(self, src):
        current_level = self._current_source_level
        current_level += 1
        max_level = self._current_source_level
        for s, n in src._unset:
            ds = self._data.get(s, None)
            if ds is not None and n in ds:
                self._data[s] = ds.preparewrite()
                del self._data[s][n]
        for s in src:
            ds = self._data.get(s, None)
            if ds:
                self._data[s] = ds.preparewrite()
            else:
                self._data[s] = util.cowsortdict()
            for k, v in src._data[s].items():
                value, source, level = v
                level += current_level
                max_level = max(level, current_level)
                self._data[s][k] = (value, source, level)
        self._current_source_level = max_level

    def _get(self, section, item):
        return self._data.get(section, {}).get(item)

    def get(self, section, item, default=None):
        result = self._get(section, item)
        if result is None:
            return default
        return result[0]

    def backup(self, section, key):
        """return a tuple allowing restore to reinstall a previous value

        The main reason we need it is because it handles the "no data" case.
        """
        try:
            item = self._data[section][key]
        except KeyError:
            return (section, key)
        else:
            return (section, key) + item

    def source(self, section, item):
        result = self._get(section, item)
        if result is None:
            return b""
        return result[1]

    def level(self, section, item):
        result = self._get(section, item)
        if result is None:
            return None
        return result[2]

    def sections(self):
        return sorted(self._data.keys())

    def items(self, section: bytes) -> list[tuple[bytes, bytes]]:
        items = self._data.get(section, {}).items()
        return [(k, v[0]) for (k, v) in items]

    def set(self, section, item, value, source=b""):
        assert not isinstance(
            section, str
        ), b'config section may not be unicode strings on Python 3'
        assert not isinstance(
            item, str
        ), b'config item may not be unicode strings on Python 3'
        assert not isinstance(
            value, str
        ), b'config values may not be unicode strings on Python 3'
        if section not in self:
            self._data[section] = util.cowsortdict()
        else:
            self._data[section] = self._data[section].preparewrite()
        self._data[section][item] = (value, source, self._current_source_level)

    def alter(self, section, key, new_value):
        """alter a value without altering its source or level

        This method is meant to be used by `ui.fixconfig` only."""
        item = self._data[section][key]
        size = len(item)
        new_item = (new_value,) + item[1:]
        assert len(new_item) == size
        self._data[section][key] = new_item

    def restore(self, data):
        """restore data returned by self.backup"""
        if len(data) != 2:
            # restore old data
            section, key = data[:2]
            item = data[2:]
            self._data[section] = self._data[section].preparewrite()
            self._data[section][key] = item
        else:
            # no data before, remove everything
            section, item = data
            if section in self._data:
                self._data[section].pop(item, None)

    def parse(
        self,
        src: bytes,
        data: bytes,
        sections: CfgSectionsT | None = None,
        remap: CfgRemapT | None = None,
        include=None,
    ) -> None:
        sectionre = util.re.compile(br'\[([^\[]+)\]')
        itemre = util.re.compile(br'([^=\s][^=]*?)\s*=\s*(.*\S|)')
        contre = util.re.compile(br'\s+(\S|\S.*\S)\s*$')
        emptyre = util.re.compile(br'(;|#|\s*$)')
        commentre = util.re.compile(br'(;|#)')
        unsetre = util.re.compile(br'%unset\s+(\S+)')
        includere = util.re.compile(br'%include\s+(\S|\S.*\S)\s*$')
        section = b""
        item = None
        line = 0
        cont = False

        if remap:
            section = remap.get(section, section)

        for l in data.splitlines(True):
            line += 1
            if line == 1 and l.startswith(b'\xef\xbb\xbf'):
                # Someone set us up the BOM
                l = l[3:]
            if cont:
                if commentre.match(l):
                    continue
                m = contre.match(l)
                if m:
                    if sections and section not in sections:
                        continue
                    v = self.get(section, item) + b"\n" + m.group(1)
                    self.set(section, item, v, b"%s:%d" % (src, line))
                    continue
                item = None
                cont = False
            m = includere.match(l)

            if m and include:
                expanded = util.expandpath(m.group(1))
                try:
                    include(expanded, remap=remap, sections=sections)
                except OSError as inst:
                    if inst.errno != errno.ENOENT:
                        raise error.ConfigError(
                            _(b"cannot include %s (%s)")
                            % (expanded, encoding.strtolocal(inst.strerror)),
                            b"%s:%d" % (src, line),
                        )
                continue
            if emptyre.match(l):
                continue
            m = sectionre.match(l)
            if m:
                section = m.group(1)
                if remap:
                    section = remap.get(section, section)
                if section not in self:
                    self._data[section] = util.cowsortdict()
                continue
            m = itemre.match(l)
            if m:
                item = m.group(1)
                cont = True
                if sections and section not in sections:
                    continue
                self.set(section, item, m.group(2), b"%s:%d" % (src, line))
                continue
            m = unsetre.match(l)
            if m:
                name = m.group(1)
                if sections and section not in sections:
                    continue
                if self.get(section, name) is not None:
                    self._data[section] = self._data[section].preparewrite()
                    del self._data[section][name]
                self._unset.append((section, name))
                continue

            message = l.rstrip()
            if l.startswith(b' '):
                message = b"unexpected leading whitespace: %s" % message
            raise error.ConfigError(message, (b"%s:%d" % (src, line)))

    def read(
        self,
        path: bytes,
        fp=None,
        sections: CfgSectionsT | None = None,
        remap: CfgRemapT | None = None,
    ) -> None:
        self.new_source()
        if not fp:
            fp = util.posixfile(path, b'rb')
        assert (
            getattr(fp, 'mode', 'rb') == 'rb'
        ), b'config files must be opened in binary mode, got fp=%r mode=%r' % (
            fp,
            fp.mode,
        )

        dir = os.path.dirname(path)

        def include(
            rel: bytes,
            remap: CfgRemapT | None,
            sections: CfgSectionsT | None,
        ) -> None:
            abs = os.path.normpath(os.path.join(dir, rel))
            self.read(abs, remap=remap, sections=sections)
            # anything after the include has a higher level
            self.new_source()

        self.parse(
            path, fp.read(), sections=sections, remap=remap, include=include
        )



# main_script - utility around the top level command for Mercurial
#
# Copyright 2005-2025 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import sys
import typing

if typing.TYPE_CHECKING:
    from ..interfaces.types import (
        RepoSetupFnT,
        RepoT,
    )


class request:
    repo: RepoT | None
    prereposetups: list[RepoSetupFnT]

    def __init__(
        self,
        args,
        ui=None,
        repo: RepoT | None = None,
        fin=None,
        fout=None,
        ferr=None,
        fmsg=None,
        prereposetups: list[RepoSetupFnT] | None = None,
    ):
        self.args = args
        self.ui = ui
        self.repo = repo

        # input/output/error streams
        self.fin = fin
        self.fout = fout
        self.ferr = ferr
        # separate stream for status/error messages
        self.fmsg = fmsg

        # remember options pre-parsed by _earlyparseopts()
        self.earlyoptions = {}

        # reposetups which run before extensions, useful for chg to pre-fill
        # low-level repo state (for example, changelog) before extensions.
        self.prereposetups = prereposetups or []

        # store the parsed and canonical command
        self.canonical_command = None

    def _runexithandlers(self) -> None:
        exc = None
        handlers = self.ui._exithandlers
        try:
            while handlers:
                func, args, kwargs = handlers.pop()
                try:
                    func(*args, **kwargs)
                except:  # re-raises below
                    if exc is None:
                        exc = sys.exc_info()[1]
                    self.ui.warnnoi18n(b'error in exit handlers:\n')
                    self.ui.traceback(force=True)
        finally:
            if exc is not None:
                raise exc



# config.py - Type annotations for config related objects
#
# Copyright 2025 Matt Harbison <mharbison72@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

from typing import (
    Mapping,
    Sequence,
)

CfgRemapT = Mapping[bytes, bytes]
"""The type for remapping one or more config section names to another."""

# TODO: is this better as a set? A custom protocol of supported ops? (bool(),
#  'x in y', possibly .add(), etc.)  Some callers pass a list, some internal
#  references are set based.
CfgSectionsT = Sequence[bytes]

# TODO: Add Protocol class for configmod.config class



from __future__ import annotations

import threading

from mercurial.node import (
    hex,
    sha1nodeconstants,
)
from mercurial import (
    mdiff,
    revlog,
)
from mercurial.revlogutils import (
    constants as revlog_constants,
)
from . import (
    basestore,
    constants,
    shallowutil,
)


class ChainIndicies:
    """A static class for easy reference to the delta chain indicies."""

    # The filename of this revision delta
    NAME = 0
    # The mercurial file node for this revision delta
    NODE = 1
    # The filename of the delta base's revision. This is useful when delta
    # between different files (like in the case of a move or copy, we can delta
    # against the original file content).
    BASENAME = 2
    # The mercurial file node for the delta base revision. This is the nullid if
    # this delta is a full text.
    BASENODE = 3
    # The actual delta or full text data.
    DATA = 4


class unioncontentstore(basestore.baseunionstore):
    def __init__(self, *stores, **kwargs):
        super().__init__(*stores, **kwargs)

        self.writestore = kwargs.get('writestore')

        # If allowincomplete==True then the union store can return partial
        # delta chains, otherwise it will throw a KeyError if a full
        # deltachain can't be found.
        self.allowincomplete = kwargs.get('allowincomplete', False)

    def get(self, name, node):
        """Fetches the full text revision contents of the given name+node pair.
        If the full text doesn't exist, throws a KeyError.

        Under the hood, this uses getdeltachain() across all the stores to build
        up a full chain to produce the full text.
        """
        chain = self.getdeltachain(name, node)

        if chain[-1][ChainIndicies.BASENODE] != sha1nodeconstants.nullid:
            # If we didn't receive a full chain, throw
            raise KeyError((name, hex(node)))

        # The last entry in the chain is a full text, so we start our delta
        # applies with that.
        fulltext = chain.pop()[ChainIndicies.DATA]

        text = fulltext
        while chain:
            delta = chain.pop()[ChainIndicies.DATA]
            text = mdiff.patches(text, [delta])

        return text

    @basestore.baseunionstore.retriable
    def getdelta(self, name, node):
        """Return the single delta entry for the given name/node pair."""
        for store in self.stores:
            try:
                return store.getdelta(name, node)
            except KeyError:
                pass

        raise KeyError((name, hex(node)))

    def getdeltachain(self, name, node):
        """Returns the deltachain for the given name/node pair.

        Returns an ordered list of:

          [(name, node, deltabasename, deltabasenode, deltacontent),...]

        where the chain is terminated by a full text entry with a nullid
        deltabasenode.
        """
        chain = self._getpartialchain(name, node)
        while chain[-1][ChainIndicies.BASENODE] != sha1nodeconstants.nullid:
            x, x, deltabasename, deltabasenode, x = chain[-1]
            try:
                morechain = self._getpartialchain(deltabasename, deltabasenode)
                chain.extend(morechain)
            except KeyError:
                # If we allow incomplete chains, don't throw.
                if not self.allowincomplete:
                    raise
                break

        return chain

    @basestore.baseunionstore.retriable
    def getmeta(self, name, node):
        """Returns the metadata dict for given node."""
        for store in self.stores:
            try:
                return store.getmeta(name, node)
            except KeyError:
                pass
        raise KeyError((name, hex(node)))

    def getmetrics(self):
        metrics = [s.getmetrics() for s in self.stores]
        return shallowutil.sumdicts(*metrics)

    @basestore.baseunionstore.retriable
    def _getpartialchain(self, name, node):
        """Returns a partial delta chain for the given name/node pair.

        A partial chain is a chain that may not be terminated in a full-text.
        """
        for store in self.stores:
            try:
                return store.getdeltachain(name, node)
            except KeyError:
                pass

        raise KeyError((name, hex(node)))

    def add(self, name, node, data):
        raise RuntimeError(
            b"cannot add content only to remotefilelog contentstore"
        )

    def getmissing(self, keys):
        missing = keys
        for store in self.stores:
            if missing:
                missing = store.getmissing(missing)
        return missing

    def addremotefilelognode(self, name, node, data):
        if self.writestore:
            self.writestore.addremotefilelognode(name, node, data)
        else:
            raise RuntimeError(b"no writable store configured")

    def markledger(self, ledger, options=None):
        for store in self.stores:
            store.markledger(ledger, options)


class remotefilelogcontentstore(basestore.basestore):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._threaddata = threading.local()

    def get(self, name, node):
        # return raw revision text
        data = self._getdata(name, node)

        offset, size, flags = shallowutil.parsesizeflags(data)
        content = data[offset : offset + size]

        ancestormap = shallowutil.ancestormap(data)
        p1, p2, linknode, copyfrom = ancestormap[node]
        copyrev = None
        if copyfrom:
            copyrev = hex(p1)

        self._updatemetacache(node, size, flags)

        # lfs tracks renames in its own metadata, remove hg copy metadata,
        # because copy metadata will be re-added by lfs flag processor.
        if flags & revlog.REVIDX_EXTSTORED:
            copyrev = copyfrom = None
        revision = shallowutil.createrevlogtext(content, copyfrom, copyrev)
        return revision

    def getdelta(self, name, node):
        # Since remotefilelog content stores only contain full texts, just
        # return that.
        revision = self.get(name, node)
        return (
            revision,
            name,
            sha1nodeconstants.nullid,
            self.getmeta(name, node),
        )

    def getdeltachain(self, name, node):
        # Since remotefilelog content stores just contain full texts, we return
        # a fake delta chain that just consists of a single full text revision.
        # The nullid in the deltabasenode slot indicates that the revision is a
        # fulltext.
        revision = self.get(name, node)
        return [(name, node, None, sha1nodeconstants.nullid, revision)]

    def getmeta(self, name, node):
        self._sanitizemetacache()
        if node != self._threaddata.metacache[0]:
            data = self._getdata(name, node)
            offset, size, flags = shallowutil.parsesizeflags(data)
            self._updatemetacache(node, size, flags)
        return self._threaddata.metacache[1]

    def add(self, name, node, data):
        raise RuntimeError(
            b"cannot add content only to remotefilelog contentstore"
        )

    def _sanitizemetacache(self):
        metacache = getattr(self._threaddata, 'metacache', None)
        if metacache is None:
            self._threaddata.metacache = (None, None)  # (node, meta)

    def _updatemetacache(self, node, size, flags):
        self._sanitizemetacache()
        if node == self._threaddata.metacache[0]:
            return
        meta = {constants.METAKEYFLAG: flags, constants.METAKEYSIZE: size}
        self._threaddata.metacache = (node, meta)


class remotecontentstore:
    def __init__(self, ui, fileservice, shared):
        self._fileservice = fileservice
        # type(shared) is usually remotefilelogcontentstore
        self._shared = shared

    def get(self, name, node):
        self._fileservice.prefetch(
            [(name, hex(node))], force=True, fetchdata=True
        )
        return self._shared.get(name, node)

    def getdelta(self, name, node):
        revision = self.get(name, node)
        return (
            revision,
            name,
            sha1nodeconstants.nullid,
            self._shared.getmeta(name, node),
        )

    def getdeltachain(self, name, node):
        # Since our remote content stores just contain full texts, we return a
        # fake delta chain that just consists of a single full text revision.
        # The nullid in the deltabasenode slot indicates that the revision is a
        # fulltext.
        revision = self.get(name, node)
        return [(name, node, None, sha1nodeconstants.nullid, revision)]

    def getmeta(self, name, node):
        self._fileservice.prefetch(
            [(name, hex(node))], force=True, fetchdata=True
        )
        return self._shared.getmeta(name, node)

    def add(self, name, node, data):
        raise RuntimeError(b"cannot add to a remote store")

    def getmissing(self, keys):
        return keys

    def markledger(self, ledger, options=None):
        pass


class manifestrevlogstore:
    _repackendlinkrev: int

    def __init__(self, repo):
        self._store = repo.store
        self._svfs = repo.svfs
        self._revlogs = dict()
        self._cl = repo.changelog
        self._repackstartlinkrev = 0

    def get(self, name, node):
        return self._revlog(name).rawdata(node)

    def getdelta(self, name, node):
        revision = self.get(name, node)
        return revision, name, self._cl.nullid, self.getmeta(name, node)

    def getdeltachain(self, name, node):
        revision = self.get(name, node)
        return [(name, node, None, self._cl.nullid, revision)]

    def getmeta(self, name, node):
        rl = self._revlog(name)
        rev = rl.rev(node)
        return {
            constants.METAKEYFLAG: rl.flags(rev),
            constants.METAKEYSIZE: rl.rawsize(rev),
        }

    def getancestors(self, name, node, known=None):
        if known is None:
            known = set()
        if node in known:
            return []

        rl = self._revlog(name)
        ancestors = {}
        missing = {node}
        for ancrev in rl.ancestors([rl.rev(node)], inclusive=True):
            ancnode = rl.node(ancrev)
            missing.discard(ancnode)

            p1, p2 = rl.parents(ancnode)
            if p1 != self._cl.nullid and p1 not in known:
                missing.add(p1)
            if p2 != self._cl.nullid and p2 not in known:
                missing.add(p2)

            linknode = self._cl.node(rl.linkrev(ancrev))
            ancestors[rl.node(ancrev)] = (p1, p2, linknode, b'')
            if not missing:
                break
        return ancestors

    def getnodeinfo(self, name, node):
        cl = self._cl
        rl = self._revlog(name)
        parents = rl.parents(node)
        linkrev = rl.linkrev(rl.rev(node))
        return (parents[0], parents[1], cl.node(linkrev), None)

    def add(self, *args):
        raise RuntimeError(b"cannot add to a revlog store")

    def _revlog(self, name) -> revlog.revlog:
        rl = self._revlogs.get(name)
        if rl is None:
            revlogname = b'00manifesttree'
            if name != b'':
                revlogname = b'meta/%s/00manifest' % name
            # XXX building your own revlog is bound to troubles
            rl = revlog.revlog(
                self._svfs,
                target=(revlog_constants.KIND_MANIFESTLOG, name),
                radix=revlogname,
            )
            self._revlogs[name] = rl
        return rl

    def getmissing(self, keys):
        missing = []
        for name, node in keys:
            mfrevlog = self._revlog(name)
            if not mfrevlog.index.has_node(node):
                missing.append((name, node))

        return missing

    def setrepacklinkrevrange(self, startrev, endrev):
        self._repackstartlinkrev = startrev
        self._repackendlinkrev = endrev

    def markledger(self, ledger, options=None):
        if options and options.get(constants.OPTION_PACKSONLY):
            return
        treename = b''
        rl = revlog.revlog(
            self._svfs,
            target=(revlog_constants.KIND_MANIFESTLOG, treename),
            radix=b'00manifesttree',
        )
        startlinkrev = self._repackstartlinkrev
        endlinkrev = self._repackendlinkrev
        for rev in range(len(rl) - 1, -1, -1):
            linkrev = rl.linkrev(rev)
            if linkrev < startlinkrev:
                break
            if linkrev > endlinkrev:
                continue
            node = rl.node(rev)
            ledger.markdataentry(self, treename, node)
            ledger.markhistoryentry(self, treename, node)

        for t, path, size in self._store.data_entries():
            if path[:5] != b'meta/' or path[-2:] != b'.i':
                continue

            treename = path[5 : -len(b'/00manifest')]

            rl = revlog.revlog(
                self._svfs,
                target=(revlog_constants.KIND_MANIFESTLOG, treename),
                radix=path[:-2],
            )
            for rev in range(len(rl) - 1, -1, -1):
                linkrev = rl.linkrev(rev)
                if linkrev < startlinkrev:
                    break
                if linkrev > endlinkrev:
                    continue
                node = rl.node(rev)
                ledger.markdataentry(self, treename, node)
                ledger.markhistoryentry(self, treename, node)

    def cleanup(self, ledger):
        pass



# bundlecaches.py - basis utility to deal with clone bundle manifests
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# Has few import on purposes, see mercurial.bundlecaches for more advanced logic.

from __future__ import annotations

CB_MANIFEST_FILE = b'clonebundles.manifest'


def get_manifest_lines(repo) -> list[bytes]:
    """get the lines of the bundle manifest to be served to a client
    from a server"""
    raw_text = repo.vfs.tryread(CB_MANIFEST_FILE)
    entries = [e.split(b' ', 1) for e in raw_text.splitlines()]

    new_lines = []
    for e in entries:
        url = alter_bundle_url(repo, e[0])
        if len(e) == 1:
            line = url + b'\n'
        else:
            line = b"%s %s\n" % (url, e[1])
        new_lines.append(line)
    return new_lines


def alter_bundle_url(repo, url: bytes) -> bytes:
    """a function that exist to help extension and hosting to alter the url

    This will typically be used to inject authentication information in the url
    of cached bundles."""
    return url



# dirstatemap.py
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import stat

from typing import (
    TYPE_CHECKING,
)

from .i18n import _
from .interfaces import (
    misc as int_misc,
)
from . import (
    error,
    pathutil,
    policy,
    testing,
    txnutil,
    typelib,
    util,
)

from .dirstateutils import (
    docket as docketmod,
    v2,
)

if TYPE_CHECKING:
    from . import (
        ui as uimod,
    )

parsers = policy.importmod('parsers')
rustmod = policy.importrust('dirstate')

propertycache = util.propertycache

if rustmod is None:
    DirstateItem = parsers.DirstateItem
else:
    DirstateItem = rustmod.DirstateItem

rangemask = 0x7FFFFFFF

WRITE_MODE_AUTO = 0
WRITE_MODE_FORCE_NEW = 1
WRITE_MODE_FORCE_APPEND = 2


V2_MAX_READ_ATTEMPTS = 5


class _dirstatemapcommon:
    """
    Methods that are identical for both implementations of the dirstatemap
    class, with and without Rust extensions enabled.
    """

    _use_dirstate_v2: bool
    _nodeconstants: typelib.NodeConstants
    _ui: uimod.ui
    _root: bytes
    _filename: bytes
    _nodelen: int
    _dirtyparents: bool
    _docket: docketmod.DirstateDocket | None
    _write_mode: int
    _pendingmode: bool | None
    identity: int_misc.ICacheStat | None

    # please pytype

    _map = None
    copymap = None

    def __init__(
        self,
        ui: uimod.ui,
        opener,
        root: bytes,
        nodeconstants: typelib.NodeConstants,
        use_dirstate_v2: bool,
    ) -> None:
        self._use_dirstate_v2 = use_dirstate_v2
        self._nodeconstants = nodeconstants
        self._ui = ui
        self._opener = opener
        self._root = root
        self._filename = b'dirstate'
        self._nodelen = 20  # Also update Rust code when changing this!
        self._parents = None
        self._dirtyparents = False
        self._docket = None
        write_mode = ui.config(b"devel", b"dirstate.v2.data_update_mode")
        if write_mode == b"auto":
            self._write_mode = WRITE_MODE_AUTO
        elif write_mode == b"force-append":
            self._write_mode = WRITE_MODE_FORCE_APPEND
        elif write_mode == b"force-new":
            self._write_mode = WRITE_MODE_FORCE_NEW
        else:
            # unknown value, fallback to default
            self._write_mode = WRITE_MODE_AUTO

        # for consistent view between _pl() and _read() invocations
        self._pendingmode = None

    def _get_current_identity(self) -> int_misc.ICacheStat | None:
        # TODO have a cleaner approach on httpstaticrepo side
        path = self._opener.join(self._filename)
        if path.startswith(b'https://') or path.startswith(b'http://'):
            return util.uncacheable_cachestat()
        try:
            return util.cachestat(path)
        except FileNotFoundError:
            return None

    def may_need_refresh(self) -> bool:
        if 'identity' not in vars(self):
            # no existing identity, we need a refresh
            return True
        if self.identity is None:
            return True
        if not self.identity.cacheable():
            # We cannot trust the entry
            # XXX this is a problem on windows, NFS, or other inode less system
            return True
        current_identity = self._get_current_identity()
        if current_identity is None:
            return True
        if not current_identity.cacheable():
            # We cannot trust the entry
            # XXX this is a problem on windows, NFS, or other inode less system
            return True
        return current_identity != self.identity

    def preload(self) -> None:
        """Loads the underlying data, if it's not already loaded"""
        self._map

    def get(self, key, default=None):
        return self._map.get(key, default)

    def __len__(self):
        return len(self._map)

    def __iter__(self):
        return iter(self._map)

    def __contains__(self, key):
        return key in self._map

    def __getitem__(self, item):
        return self._map[item]

    ### disk interaction

    def _opendirstatefile(self):
        fp, mode = txnutil.trypending(self._root, self._opener, self._filename)
        if self._pendingmode is not None and self._pendingmode != mode:
            fp.close()
            raise error.Abort(
                _(b'working directory state may be changed parallelly')
            )
        self._pendingmode = mode
        return fp

    def _readdirstatefile(
        self,
        size: int = -1,
    ) -> tuple[int_misc.ICacheStat | None, bytes]:
        """read the content of the file used as "entry point" for the dirstate

        Return a (identity, data) tuple. The identity can be used for cache
        validation and concurrent changes detection and must be set as
        `self.identity` if `data` is preserved.
        """
        identity = self._get_current_identity()
        # There is a race condition between fetching the identity and reading
        # the file content.  Another process might update the file after we get
        # the identity information.  However this is fine for our purpose as
        # this will only create false-positive for "data changed" and no
        # false-negative.
        #
        # in addition for the case that matter the most (updating the dirstate
        # semantic content and not just some cache information), this will not
        # happens as the lock should be held when changes to the dirstate
        # content are made.
        testing.wait_on_cfg(self._ui, b'dirstate.pre-read-file')
        try:
            with self._opendirstatefile() as fp:
                data = fp.read(size)
        except FileNotFoundError:
            # File doesn't exist, so the current state is empty
            data = b''
        testing.wait_on_cfg(self._ui, b'dirstate.post-docket-read-file')
        return identity, data

    @property
    def docket(self) -> docketmod.DirstateDocket:
        if not self._docket:
            if not self._use_dirstate_v2:
                raise error.ProgrammingError(
                    b'dirstate only has a docket in v2 format'
                )
            self.identity, data = self._readdirstatefile()
            if data == b'' or data.startswith(docketmod.V2_FORMAT_MARKER):
                self._docket = docketmod.DirstateDocket.parse(
                    data, self._nodeconstants
                )
            else:
                raise error.CorruptedDirstate(b"dirstate is not in v2 format")
        return self._docket

    def _read_v2_data(self):
        attempts = 0
        while attempts < V2_MAX_READ_ATTEMPTS:
            attempts += 1
            try:
                # TODO: use mmap when possible
                return self._opener.read(self.docket.data_filename())
            except FileNotFoundError:
                # read race detected between docket and data file
                # reload the docket and retry
                self._docket = None
        msg = b"dirstate read race happened %d times in a row"
        msg %= attempts
        raise error.Abort(msg)

    def write_v2_no_append(self, tr, st, meta, packed):
        try:
            old_docket = self.docket
        except error.CorruptedDirstate:
            # This means we've identified a dirstate-v1 file on-disk when we
            # were expecting a dirstate-v2 docket. We've managed to recover
            # from that unexpected situation, and now we want to write back a
            # dirstate-v2 file to make the on-disk situation right again.
            #
            # This shouldn't be triggered since `self.docket` is cached and
            # we would have called parents() or read() first, but it's here
            # just in case.
            old_docket = None

        new_docket = docketmod.DirstateDocket.with_new_uuid(
            self.parents(), len(packed), meta
        )
        if old_docket is not None and old_docket.uuid == new_docket.uuid:
            raise error.ProgrammingError(b'dirstate docket name collision')
        data_filename = new_docket.data_filename()
        self._opener.write(data_filename, packed)
        # tell the transaction that we are adding a new file
        if tr is not None:
            tr.addbackup(data_filename, location=b'plain')
        # Write the new docket after the new data file has been
        # written. Because `st` was opened with `atomictemp=True`,
        # the actual `.hg/dirstate` file is only affected on close.
        st.write(new_docket.serialize())
        st.close()
        # Remove the old data file after the new docket pointing to
        # the new data file was written.
        if old_docket is not None and old_docket.uuid:
            data_filename = old_docket.data_filename()
            if tr is not None:
                tr.addbackup(data_filename, location=b'plain')
            unlink = lambda _tr=None: self._opener.unlink(data_filename)
            if tr:
                category = b"dirstate-v2-clean-" + old_docket.uuid
                tr.addpostclose(category, unlink)
            else:
                unlink()
        self._docket = new_docket

    ### reading/setting parents

    def parents(self):
        if not self._parents:
            if self._use_dirstate_v2:
                try:
                    self.docket
                except error.CorruptedDirstate as e:
                    # fall back to dirstate-v1 if we fail to read v2
                    self._v1_parents(e)
                else:
                    self._parents = self.docket.parents
            else:
                self._v1_parents()

        return self._parents

    def _v1_parents(self, from_v2_exception=None):
        read_len = self._nodelen * 2
        _identity, st = self._readdirstatefile(read_len)
        l = len(st)
        if l == read_len:
            self._parents = (
                st[: self._nodelen],
                st[self._nodelen : 2 * self._nodelen],
            )
        elif l == 0:
            self._parents = (
                self._nodeconstants.nullid,
                self._nodeconstants.nullid,
            )
        else:
            hint = None
            if from_v2_exception is not None:
                hint = _(b"falling back to dirstate-v1 from v2 also failed")
            raise error.Abort(
                _(b'working directory state appears damaged!'), hint
            )


class dirstatemap(_dirstatemapcommon):
    """Map encapsulating the dirstate's contents.

    The dirstate contains the following state:

    - `identity` is the identity of the dirstate file, which can be used to
      detect when changes have occurred to the dirstate file.

    - `parents` is a pair containing the parents of the working copy. The
      parents are updated by calling `setparents`.

    - the state map maps filenames to tuples of (state, mode, size, mtime),
      where state is a single character representing 'normal', 'added',
      'removed', or 'merged'. It is read by treating the dirstate as a
      dict.  File state is updated by calling various methods (see each
      documentation for details):

      - `reset_state`,
      - `set_tracked`
      - `set_untracked`
      - `set_clean`
      - `set_possibly_dirty`

    - `copymap` maps destination filenames to their source filename.

    The dirstate also provides the following views onto the state:

    - `filefoldmap` is a dict mapping normalized filenames to the denormalized
      form that they appear as in the dirstate.

    - `dirfoldmap` is a dict mapping normalized directory names to the
      denormalized form that they appear as in the dirstate.
    """

    ### Core data storage and access

    @propertycache
    def _map(self):
        self._map = {}
        self.read()
        return self._map

    @propertycache
    def copymap(self):
        self.copymap = {}
        self._map
        return self.copymap

    def clear(self):
        self._map.clear()
        self.copymap.clear()
        self.setparents(self._nodeconstants.nullid, self._nodeconstants.nullid)
        util.clearcachedproperty(self, b"_dirs")
        util.clearcachedproperty(self, b"_alldirs")
        util.clearcachedproperty(self, b"filefoldmap")
        util.clearcachedproperty(self, b"dirfoldmap")

    def items(self):
        return self._map.items()

    # forward for python2,3 compat
    iteritems = items

    def debug_iter(self, all):
        """
        Return an iterator of (filename, state, mode, size, mtime) tuples

        `all` is unused when Rust is not enabled
        """
        for filename, item in self.items():
            yield (filename, item.state, item.mode, item.size, item.mtime)

    def keys(self):
        return self._map.keys()

    ### reading/setting parents

    def setparents(self, p1, p2, fold_p2=False):
        self._parents = (p1, p2)
        self._dirtyparents = True
        copies = {}
        if fold_p2:
            for f, s in self._map.items():
                # Discard "merged" markers when moving away from a merge state
                if s.p2_info:
                    source = self.copymap.pop(f, None)
                    if source:
                        copies[f] = source
                    s.drop_merge_data()
        return copies

    ### disk interaction

    def read(self):
        if self._use_dirstate_v2:
            try:
                self.docket
            except error.CorruptedDirstate:
                # fall back to dirstate-v1 if we fail to read v2
                self.identity, st = self._readdirstatefile()
            else:
                if not self.docket.uuid:
                    return
                st = self._read_v2_data()
        else:
            self.identity, st = self._readdirstatefile()

        if not st:
            return

        # TODO: adjust this estimate for dirstate-v2
        if hasattr(parsers, 'dict_new_presized'):
            # Make an estimate of the number of files in the dirstate based on
            # its size. This trades wasting some memory for avoiding costly
            # resizes. Each entry have a prefix of 17 bytes followed by one or
            # two path names. Studies on various large-scale real-world repositories
            # found 54 bytes a reasonable upper limit for the average path names.
            # Copy entries are ignored for the sake of this estimate.
            self._map = parsers.dict_new_presized(len(st) // 71)

        # Python's garbage collector triggers a GC each time a certain number
        # of container objects (the number being defined by
        # gc.get_threshold()) are allocated. parse_dirstate creates a tuple
        # for each file in the dirstate. The C version then immediately marks
        # them as not to be tracked by the collector. However, this has no
        # effect on when GCs are triggered, only on what objects the GC looks
        # into. This means that O(number of files) GCs are unavoidable.
        # Depending on when in the process's lifetime the dirstate is parsed,
        # this can get very expensive. As a workaround, disable GC while
        # parsing the dirstate.
        #
        # (we cannot decorate the function directly since it is in a C module)
        if self._use_dirstate_v2:
            try:
                self.docket
            except error.CorruptedDirstate:
                # fall back to dirstate-v1 if we fail to parse v2
                parse_dirstate = util.nogc(parsers.parse_dirstate)
                p = parse_dirstate(self._map, self.copymap, st)
            else:
                p = self.docket.parents
                meta = self.docket.tree_metadata
                parse_dirstate = util.nogc(v2.parse_dirstate)
                parse_dirstate(self._map, self.copymap, st, meta)
        else:
            parse_dirstate = util.nogc(parsers.parse_dirstate)
            p = parse_dirstate(self._map, self.copymap, st)
        if not self._dirtyparents:
            self.setparents(*p)

        # Avoid excess attribute lookups by fast pathing certain checks
        self.__contains__ = self._map.__contains__
        self.__getitem__ = self._map.__getitem__
        self.get = self._map.get

    def write(self, tr, st):
        if self._use_dirstate_v2:
            packed, meta = v2.pack_dirstate(self._map, self.copymap)
            self.write_v2_no_append(tr, st, meta, packed)
        else:
            packed = parsers.pack_dirstate(
                self._map, self.copymap, self.parents()
            )
            st.write(packed)
            st.close()
        self._dirtyparents = False

    @propertycache
    def identity(self) -> int_misc.ICacheStat | None:
        """A cache identifier for the state of the file as data were read

        This must always be set with the object returned from
        `self._readdirstatefile()`. assigning another value later will break
        some security mechanism and can lead to misbehavior when concurrent
        operation are run.
        """
        self._map
        return self.identity

    ### code related to maintaining and accessing "extra" property
    # (e.g. "has_dir")

    def _dirs_incr(self, filename, old_entry=None):
        """increment the dirstate counter if applicable"""
        if (
            old_entry is None or old_entry.removed
        ) and "_dirs" in self.__dict__:
            self._dirs.addpath(filename)
        if old_entry is None and "_alldirs" in self.__dict__:
            self._alldirs.addpath(filename)

    def _dirs_decr(self, filename, old_entry=None, remove_variant=False):
        """decrement the dirstate counter if applicable"""
        if old_entry is not None:
            if "_dirs" in self.__dict__ and not old_entry.removed:
                self._dirs.delpath(filename)
            if "_alldirs" in self.__dict__ and not remove_variant:
                self._alldirs.delpath(filename)
        elif remove_variant and "_alldirs" in self.__dict__:
            self._alldirs.addpath(filename)
        if "filefoldmap" in self.__dict__:
            normed = util.normcase(filename)
            self.filefoldmap.pop(normed, None)

    @propertycache
    def filefoldmap(self):
        """Returns a dictionary mapping normalized case paths to their
        non-normalized versions.
        """
        try:
            makefilefoldmap = parsers.make_file_foldmap
        except AttributeError:
            pass
        else:
            return makefilefoldmap(
                self._map, util.normcasespec, util.normcasefallback
            )

        f = {}
        normcase = util.normcase
        for name, s in self._map.items():
            if not s.removed:
                f[normcase(name)] = name
        f[b'.'] = b'.'  # prevents useless util.fspath() invocation
        return f

    @propertycache
    def dirfoldmap(self):
        f = {}
        normcase = util.normcase
        for name in self._dirs:
            f[normcase(name)] = name
        return f

    def hastrackeddir(self, d):
        """
        Returns True if the dirstate contains a tracked (not removed) file
        in this directory.
        """
        return d in self._dirs

    def hasdir(self, d):
        """
        Returns True if the dirstate contains a file (tracked or removed)
        in this directory.
        """
        return d in self._alldirs

    @propertycache
    def _dirs(self):
        return pathutil.dirs(self._map, only_tracked=True)

    @propertycache
    def _alldirs(self):
        return pathutil.dirs(self._map)

    ### code related to manipulation of entries and copy-sources

    def reset_state(
        self,
        filename,
        wc_tracked=False,
        p1_tracked=False,
        p2_info=False,
        has_meaningful_mtime=True,
        parentfiledata=None,
    ):
        """Set a entry to a given state, diregarding all previous state

        This is to be used by the part of the dirstate API dedicated to
        adjusting the dirstate after a update/merge.

        note: calling this might result to no entry existing at all if the
        dirstate map does not see any point at having one for this file
        anymore.
        """
        # copy information are now outdated
        # (maybe new information should be in directly passed to this function)
        self.copymap.pop(filename, None)

        if not (p1_tracked or p2_info or wc_tracked):
            old_entry = self._map.get(filename)
            self._drop_entry(filename)
            self._dirs_decr(filename, old_entry=old_entry)
            return

        old_entry = self._map.get(filename)
        self._dirs_incr(filename, old_entry)
        entry = DirstateItem(
            wc_tracked=wc_tracked,
            p1_tracked=p1_tracked,
            p2_info=p2_info,
            has_meaningful_mtime=has_meaningful_mtime,
            parentfiledata=parentfiledata,
        )
        self._map[filename] = entry

    def set_tracked(self, filename):
        new = False
        entry = self.get(filename)
        if entry is None:
            self._dirs_incr(filename)
            entry = DirstateItem(
                wc_tracked=True,
            )

            self._map[filename] = entry
            new = True
        elif not entry.tracked:
            self._dirs_incr(filename, entry)
            entry.set_tracked()
            self._refresh_entry(filename, entry)
            new = True
        else:
            # XXX This is probably overkill for more case, but we need this to
            # fully replace the `normallookup` call with `set_tracked` one.
            # Consider smoothing this in the future.
            entry.set_possibly_dirty()
            self._refresh_entry(filename, entry)
        return new

    def set_untracked(self, f):
        """Mark a file as no longer tracked in the dirstate map"""
        entry = self.get(f)
        if entry is None:
            return False
        else:
            self._dirs_decr(f, old_entry=entry, remove_variant=not entry.added)
            if not entry.p2_info:
                self.copymap.pop(f, None)
            entry.set_untracked()
            self._refresh_entry(f, entry)
            return True

    def set_clean(self, filename, mode, size, mtime):
        """mark a file as back to a clean state"""
        entry = self[filename]
        size = size & rangemask
        entry.set_clean(mode, size, mtime)
        self._refresh_entry(filename, entry)
        self.copymap.pop(filename, None)

    def set_possibly_dirty(self, filename):
        """record that the current state of the file on disk is unknown"""
        entry = self[filename]
        entry.set_possibly_dirty()
        self._refresh_entry(filename, entry)

    def _refresh_entry(self, f, entry):
        """record updated state of an entry"""
        if not entry.any_tracked:
            self._map.pop(f, None)

    def _drop_entry(self, f):
        """remove any entry for file f

        This should also drop associated copy information

        The fact we actually need to drop it is the responsability of the caller
        """
        self._map.pop(f, None)
        self.copymap.pop(f, None)


if rustmod is not None:

    class dirstatemap(_dirstatemapcommon):
        ### Core data storage and access

        @propertycache
        def _map(self):
            """
            Fills the Dirstatemap when called.
            """
            # ignore HG_PENDING because identity is used only for writing

            if self._use_dirstate_v2:
                try:
                    self.docket
                except error.CorruptedDirstate as e:
                    # fall back to dirstate-v1 if we fail to read v2
                    parents = self._v1_map(e)
                else:
                    parents = self.docket.parents
                    identity = self._get_rust_identity()
                    if not self.docket.uuid:
                        data = b''
                        self._map = rustmod.DirstateMap.new_empty()
                    else:
                        data = self._read_v2_data()
                        self._map = rustmod.DirstateMap.new_v2(
                            data,
                            self.docket.data_size,
                            self.docket.tree_metadata,
                            self.docket.uuid,
                            identity,
                        )
                    parents = self.docket.parents
            else:
                parents = self._v1_map()

            if parents and not self._dirtyparents:
                self.setparents(*parents)

            self.__contains__ = self._map.__contains__
            self.__getitem__ = self._map.__getitem__
            self.get = self._map.get
            return self._map

        def _get_rust_identity(self):
            identity = None
            if self.identity is not None and self.identity.stat is not None:
                stat_info = self.identity.stat
                identity = rustmod.DirstateIdentity(
                    mode=stat_info.st_mode,
                    dev=stat_info.st_dev,
                    ino=stat_info.st_ino,
                    nlink=stat_info.st_nlink,
                    uid=stat_info.st_uid,
                    gid=stat_info.st_gid,
                    size=stat_info.st_size,
                    mtime=stat_info[stat.ST_MTIME],
                    mtime_nsec=0,
                    ctime=stat_info[stat.ST_CTIME],
                    ctime_nsec=0,
                )
            return identity

        def _v1_map(self, from_v2_exception=None):
            try:
                self.identity, data = self._readdirstatefile()
                identity = self._get_rust_identity()
                self._map, parents = rustmod.DirstateMap.new_v1(data, identity)
            except OSError as e:
                if from_v2_exception is not None:
                    raise e from from_v2_exception
                raise
            return parents

        @property
        def copymap(self):
            return self._map.copymap()

        def debug_iter(self, all):
            """
            Return an iterator of (filename, state, mode, size, mtime) tuples

            `all`: also include with `state == b' '` dirstate tree nodes that
            don't have an associated `DirstateItem`.

            """
            return self._map.debug_iter(all)

        def clear(self):
            self._map.clear()
            self.setparents(
                self._nodeconstants.nullid, self._nodeconstants.nullid
            )
            util.clearcachedproperty(self, b"_dirs")
            util.clearcachedproperty(self, b"_alldirs")
            util.clearcachedproperty(self, b"dirfoldmap")

        def items(self):
            return self._map.items()

        # forward for python2,3 compat
        iteritems = items

        def keys(self):
            return iter(self._map)

        ### reading/setting parents

        def setparents(self, p1, p2, fold_p2=False):
            self._parents = (p1, p2)
            self._dirtyparents = True
            copies = {}
            if fold_p2:
                copies = self._map.setparents_fixup()
            return copies

        ### disk interaction

        @propertycache
        def identity(self):
            self._map
            return self.identity

        def write(self, tr, st):
            if not self._use_dirstate_v2:
                p1, p2 = self.parents()
                packed = self._map.write_v1(p1, p2)
                st.write(packed)
                st.close()
                self._dirtyparents = False
                return

            write_mode = self._write_mode
            try:
                docket = self.docket
            except error.CorruptedDirstate:
                # fall back to dirstate-v1 if we fail to parse v2
                docket = None

            # We can only append to an existing data file if there is one
            if docket is None or docket.uuid is None:
                write_mode = WRITE_MODE_FORCE_NEW
            packed, meta, append = self._map.write_v2(write_mode)
            if append:
                docket = self.docket
                data_filename = docket.data_filename()
                # We mark it for backup to make sure a future `hg rollback` (or
                # `hg recover`?) call find the data it needs to restore a
                # working repository.
                #
                # The backup can use a hardlink because the format is resistant
                # to trailing "dead" data.
                if tr is not None:
                    tr.addbackup(data_filename, location=b'plain')
                with self._opener(data_filename, b'r+b') as fp:
                    fp.seek(docket.data_size)
                    assert fp.tell() == docket.data_size
                    written = fp.write(packed)
                    if written is not None:  # py2 may return None
                        assert written == len(packed), (written, len(packed))
                docket.data_size += len(packed)
                docket.parents = self.parents()
                docket.tree_metadata = meta
                st.write(docket.serialize())
                st.close()
            else:
                self.write_v2_no_append(tr, st, meta, packed)
            # Reload from the newly-written file
            util.clearcachedproperty(self, b"_map")
            self._dirtyparents = False

        ### code related to maintaining and accessing "extra" property
        # (e.g. "has_dir")

        @propertycache
        def filefoldmap(self):
            """Returns a dictionary mapping normalized case paths to their
            non-normalized versions.
            """
            return self._map.filefoldmapasdict()

        def hastrackeddir(self, d):
            return self._map.hastrackeddir(d)

        def hasdir(self, d):
            return self._map.hasdir(d)

        @propertycache
        def dirfoldmap(self):
            f = {}
            normcase = util.normcase
            for name in self._map.tracked_dirs():
                f[normcase(name)] = name
            return f

        ### code related to manipulation of entries and copy-sources

        def set_tracked(self, f):
            return self._map.set_tracked(f)

        def set_untracked(self, f):
            return self._map.set_untracked(f)

        def set_clean(self, filename, mode, size, mtime):
            self._map.set_clean(filename, mode, size, mtime)

        def set_possibly_dirty(self, f):
            self._map.set_possibly_dirty(f)

        def reset_state(
            self,
            filename,
            wc_tracked=False,
            p1_tracked=False,
            p2_info=False,
            has_meaningful_mtime=True,
            parentfiledata=None,
        ):
            return self._map.reset_state(
                filename,
                wc_tracked,
                p1_tracked,
                p2_info,
                has_meaningful_mtime,
                parentfiledata,
            )
